# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Admin model views for PersistentIdentifier."""

import uuid

from flask import current_app, flash, url_for
from flask_admin.actions import action
from flask_admin.contrib.sqla import ModelView
from flask_wtf import FlaskForm
from invenio_admin.filters import FilterConverter
from invenio_admin.forms import LazyChoices
from markupsafe import Markup
from wtforms.validators import ValidationError

from .models import (
    Bucket,
    FileInstance,
    Location,
    MultipartObject,
    ObjectVersion,
    slug_pattern,
)
from .tasks import verify_checksum


def _(x):
    """Identity function for string extraction."""
    return x


def require_slug(form, field):
    """Validate location name."""
    if not slug_pattern.match(field.data):
        raise ValidationError(_("Invalid location name."))


def link(text, link_func):
    """Generate a object formatter for links.."""

    def object_formatter(v, c, m, p):
        """Format object view link."""
        return Markup('<a href="{0}">{1}</a>'.format(link_func(m), text))

    return object_formatter


class LocationModelView(ModelView):
    """ModelView for the locations."""

    filter_converter = FilterConverter()
    can_create = True
    can_edit = False
    can_delete = True
    can_view_details = True
    column_formatters = dict(
        buckets=link("Buckets", lambda o: url_for("bucket.index_view", flt2_2=o.name))
    )
    column_details_list = ("name", "uri", "default", "created", "updated", "buckets")
    column_list = ("name", "uri", "default", "created", "updated", "buckets")
    column_labels = dict(
        id=_("ID"),
        uri=_("URI"),
    )
    column_filters = (
        "default",
        "created",
        "updated",
    )
    column_searchable_list = ("uri", "name")
    column_default_sort = "name"
    form_base_class = FlaskForm
    form_columns = ("name", "uri", "default")
    form_args = dict(name=dict(validators=[require_slug]))
    page_size = 25


class BucketModelView(ModelView):
    """ModelView for the buckets."""

    filter_converter = FilterConverter()
    can_create = True
    can_delete = False
    can_edit = True
    can_view_details = True
    column_formatters = dict(
        objects=link(
            "Objects",
            lambda o: url_for(
                "objectversion.index_view", flt0_0=o.id, flt1_37=1, sort=1
            ),
        ),
        object_versions=link(
            "Object Versions",
            lambda o: url_for(
                "objectversion.index_view", flt0_0=o.id, flt1_29=1, sort=1
            ),
        ),
    )
    column_details_list = (
        "id",
        "location",
        "default_storage_class",
        "deleted",
        "locked",
        "size",
        "quota_size",
        "max_file_size",
        "created",
        "updated",
        "objects",
        "object_versions",
    )
    column_list = (
        "id",
        "location",
        "default_storage_class",
        "deleted",
        "locked",
        "size",
        "quota_size",
        "created",
        "updated",
        "objects",
    )
    column_labels = dict(
        id=_("UUID"),
        default_location=_("Location"),
        pid_provider=_("Storage Class"),
    )
    column_filters = (
        # Change of order affects Location.column_formatters!
        "location.name",
        "default_storage_class",
        "deleted",
        "locked",
        "size",
        "created",
        "updated",
    )
    column_default_sort = ("updated", True)
    form_base_class = FlaskForm
    form_columns = (
        "default_storage_class",
        "locked",
        "deleted",
        "quota_size",
        "max_file_size",
    )
    form_choices = dict(
        default_storage_class=LazyChoices(
            lambda: current_app.config["FILES_REST_STORAGE_CLASS_LIST"].items()
        )
    )
    page_size = 25


class ObjectModelView(ModelView):
    """ModelView for the objects."""

    filter_converter = FilterConverter()
    can_create = False
    can_edit = False
    can_delete = False
    can_view_details = True
    column_formatters = dict(
        file_instance=link(
            "File", lambda o: url_for("fileinstance.index_view", flt0_0=o.file_id)
        ),
        versions=link(
            "Versions",
            lambda o: url_for(
                "objectversion.index_view",
                sort=7,
                desc=1,
                flt0_0=o.bucket_id,
                flt1_29=o.key,
            ),
        ),
        bucket_objs=link(
            "Objects",
            lambda o: url_for(
                "objectversion.index_view", flt0_0=o.bucket_id, flt1_37=1, sort=1
            ),
        ),
    )
    column_labels = {
        "version_id": _("Version"),
        "file_id": _("File UUID"),
        "file.uri": _("URI"),
        "file.size": _("Size"),
        "is_deleted": _("Deleted"),
        "file.checksum": _("Checksum"),
        "file.readable": _("Readable"),
        "file.storage_class": _("Storage class"),
        "bucket_objs": _("Objects"),
        "file_instance": _("File"),
    }
    column_list = (
        "bucket",
        "key",
        "version_id",
        "file.uri",
        "is_head",
        "is_deleted",
        "file.size",
        "created",
        "updated",
        "versions",
        "bucket_objs",
        "file_instance",
    )
    column_searchable_list = ("key",)
    column_details_list = (
        "bucket",
        "key",
        "version_id",
        "file_id",
        "file.uri",
        "file.checksum",
        "file.size",
        "file.storage_class",
        "is_head",
        "is_deleted",
        "created",
        "updated",
        "file_instance",
        "versions",
    )
    column_filters = (
        # Order affects column_formatters in other model views.
        "bucket.id",
        "bucket.locked",
        "bucket.deleted",
        "bucket.location.name",
        "file_id",
        "file.checksum",
        "file.storage_class",
        "file.readable",
        "key",
        "version_id",
        "is_head",
        "file.size",
        "created",
        "updated",
    )
    column_sortable_list = (
        "key",
        "file.uri",
        "is_head",
        "file.size",
        "created",
        "updated",
    )
    column_default_sort = ("updated", True)
    page_size = 25


class FileInstanceModelView(ModelView):
    """ModelView for the objects."""

    filter_converter = FilterConverter()
    can_create = False
    can_edit = False
    can_delete = False
    can_view_details = True
    column_formatters = dict(
        objects=link(
            "Objects", lambda o: url_for("objectversion.index_view", flt3_12=o.id)
        ),
    )
    column_labels = dict(
        id=_("ID"),
        uri=_("URI"),
        last_check=_("Fixity"),
        last_check_at=_("Checked"),
    )
    column_list = (
        "id",
        "uri",
        "storage_class",
        "size",
        "checksum",
        "readable",
        "writable",
        "last_check",
        "last_check_at",
        "created",
        "updated",
        "objects",
    )
    column_searchable_list = (
        "uri",
        "size",
        "checksum",
    )
    column_details_list = (
        "id",
        "uri",
        "storage_class",
        "size",
        "checksum",
        "readable",
        "writable",
        "last_check",
        "last_check_at",
        "created",
        "updated",
        "objects",
    )
    column_filters = (
        "id",
        "uri",
        "storage_class",
        "size",
        "checksum",
        "readable",
        "writable",
        "last_check",
        "last_check_at",
        "created",
        "updated",
        "objects",
    )
    column_default_sort = ("updated", True)
    page_size = 25

    @action("verify_checksum", _("Run fixity check"))
    def action_verify_checksum(self, ids):
        """Inactivate users."""
        try:
            count = 0
            for file_id in ids:
                f = FileInstance.query.filter_by(id=uuid.UUID(file_id)).one_or_none()
                if f is None:
                    raise ValueError(_("Cannot find file instance."))
                verify_checksum.delay(file_id)
                count += 1
            if count > 0:
                flash(_("Fixity check(s) sent to queue."), "success")
        except Exception as exc:
            if not self.handle_view_exception(exc):
                raise
            current_app.logger.exception(str(exc))  # pragma: no cover
            flash(_("Failed to run fixity checks."), "error")  # pragma: no cover


class MultipartObjectModelView(ModelView):
    """ModelView for the objects."""

    filter_converter = FilterConverter()
    can_create = False
    can_edit = False
    can_delete = False
    can_view_details = True
    column_formatters = dict(
        file_instance=link(
            "File", lambda o: url_for("fileinstance.index_view", flt0_0=o.file_id)
        ),
    )
    column_labels = dict(
        id=_("ID"),
        completed=_("Complete"),
        file_instance=_("File"),
    )
    column_list = (
        "upload_id",
        "completed",
        "created",
        "updated",
        "file_instance",
    )
    column_details_list = (
        "upload_id",
        "completed",
        "created",
        "updated",
        "file_instance",
    )
    column_filters = (
        "upload_id",
        "completed",
        "created",
        "updated",
    )
    column_default_sort = ("upload_id", True)
    page_size = 25


location_adminview = dict(
    modelview=LocationModelView, model=Location, category=_("Files")
)
bucket_adminview = dict(modelview=BucketModelView, model=Bucket, category=_("Files"))
object_adminview = dict(
    modelview=ObjectModelView, model=ObjectVersion, category=_("Files")
)
fileinstance_adminview = dict(
    modelview=FileInstanceModelView, model=FileInstance, category=_("Files")
)
multipartobject_adminview = dict(
    modelview=MultipartObjectModelView, model=MultipartObject, category=_("Files")
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Request customization tests."""

from io import BytesIO

from flask import request

from invenio_files_rest.app import Flask


def test_max_content_length():
    """Test max content length."""
    max_len = 10

    app = Flask("test")
    app.config["MAX_CONTENT_LENGTH"] = max_len

    @app.route("/test", methods=["GET", "PUT"])
    def test():
        # Access request.form to ensure form parser kicks-in.
        request.form.to_dict()
        return request.stream.read()

    with app.test_client() as client:
        # No content-type, no max content length checking
        data = b"a" * (max_len + 1)
        res = client.put("/test", input_stream=BytesIO(data))
        assert res.status_code == 200
        assert res.data == data

        # Non-formdata content-type, no max content length checking
        res = client.put(
            "/test",
            input_stream=BytesIO(data),
            headers={"Content-Type": "application/octet-stream"},
        )
        assert res.status_code == 200
        assert res.data == data

        # With form data content-type (below max content length)
        res = client.put(
            "/test", data={"123": "a" * (max_len - 4)}  # content-length == 10
        )
        assert res.status_code == 200
        # Because formdata parsing reads the data stream, trying to read it
        # again means we just get empty data:
        assert res.data == b""

        # With formdata content-type (above max content length)
        res = client.put(
            "/test", data={"123": "a" * (max_len - 3)}  # content-length == 11
        )
        assert res.status_code == 413



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module test views."""

import json
from io import BytesIO
from unittest.mock import MagicMock, patch

import pytest
from flask import url_for
from testutils import BadBytesIO, login_user

from invenio_files_rest.models import Bucket, MultipartObject, Part
from invenio_files_rest.tasks import merge_multipartobject


def obj_url(bucket):
    """Get object URL."""
    return url_for(
        "invenio_files_rest.object_api",
        bucket_id=str(bucket.id),
        key="mybigfile",
    )


@pytest.mark.parametrize(
    "user, expected",
    [
        (None, 404),
        ("auth", 404),
        ("objects", 404),  # TODO - use 403 instead
        ("bucket", 200),
        ("location", 200),
    ],
)
def test_post_init(client, headers, permissions, bucket, get_json, user, expected):
    """Test init multipart upload."""
    login_user(client, permissions[user])

    # Initiate multipart upload
    res = client.post(
        obj_url(bucket),
        query_string="uploads",
        headers={"Content-Type": "application/json"},
        data=json.dumps(
            {
                "size": 10,
                "partSize": 4,
            }
        ),
    )
    assert res.status_code == expected

    if res.status_code == 200:
        data = get_json(res)
        expected_keys = [
            "id",
            "bucket",
            "completed",
            "size",
            "part_size",
            "last_part_number",
            "last_part_size",
            "links",
        ]
        for k in expected_keys:
            assert k in data


def test_post_init_querystring(client, bucket, get_json, admin_user):
    """Test init multipart upload."""
    login_user(client, admin_user)

    res = client.post(
        obj_url(bucket),
        query_string="uploads&size=10&partSize=4",
    )
    assert res.status_code == 200


def test_get_init_not_allowed(client, bucket, get_json):
    """Test get request to ?uploads."""
    res = client.get(
        obj_url(bucket),
        query_string="uploads",
    )
    assert res.status_code == 405


def test_post_invalid_partsizes(client, headers, bucket, get_json, admin_user):
    """Test invalid multipart init."""
    login_user(client, admin_user)

    # Part size too large
    res = client.post(
        obj_url(bucket),
        query_string="uploads",
        headers=headers,
        data=json.dumps({"size": 30, "partSize": 21}),
    )
    assert res.status_code == 400

    # Part size too small
    res = client.post(
        obj_url(bucket),
        query_string="uploads",
        headers=headers,
        data=json.dumps({"size": 30, "partSize": 1}),
    )
    assert res.status_code == 400

    # Size too large
    res = client.post(
        obj_url(bucket),
        query_string="uploads",
        headers=headers,
        data=json.dumps({"size": 2 * 100 + 1, "partSize": 2}),
    )
    assert res.status_code == 400


def test_post_size_limits(client, db, headers, bucket, admin_user):
    """Test invalid multipart init."""
    login_user(client, admin_user)

    bucket.quota_size = 100
    db.session.commit()

    # Bucket quota exceed
    res = client.post(
        obj_url(bucket),
        query_string="uploads",
        headers=headers,
        data=json.dumps({"size": 101, "partSize": 20}),
    )
    assert res.status_code == 400

    bucket.max_file_size = 50
    db.session.commit()

    # Max file size exceeded
    res = client.post(
        obj_url(bucket),
        query_string="uploads",
        headers=headers,
        data=json.dumps({"size": 51, "partSize": 20}),
    )
    assert res.status_code == 400


def test_post_locked_bucket(client, db, headers, bucket, get_json, admin_user):
    """Test invalid multipart init."""
    login_user(client, admin_user)

    bucket.locked = True
    db.session.commit()

    res = client.post(
        obj_url(bucket),
        query_string="uploads",
        headers=headers,
        data=json.dumps({"size": 10, "partSize": 2}),
    )
    assert res.status_code == 403

    bucket.deleted = True
    db.session.commit()

    res = client.post(
        obj_url(bucket),
        query_string="uploads",
        headers=headers,
        data=json.dumps({"size": 10, "partSize": 2}),
    )
    assert res.status_code == 404


def test_post_invalidkey(client, db, headers, bucket, admin_user):
    """Test invalid multipart init."""
    login_user(client, admin_user)

    object_url = (
        url_for(
            "invenio_files_rest.object_api",
            bucket_id=str(bucket.id),
            key="a" * 1025,
        )
        + "uploads"
    )

    # Bucket quota exceed
    res = client.post(
        object_url,
        query_string="uploads",
        headers=headers,
        data=json.dumps({"size": 50, "partSize": 20}),
    )
    assert res.status_code == 400


@pytest.mark.parametrize(
    "user, expected",
    [
        (None, 404),
        ("auth", 404),
        ("objects", 404),  # TODO - use 403 instead
        ("bucket", 200),
        ("location", 200),
    ],
)
def test_put(
    client,
    db,
    bucket,
    permissions,
    multipart,
    multipart_url,
    get_md5,
    get_json,
    user,
    expected,
):
    """Test part upload."""
    login_user(client, permissions[user])

    data = b"a" * multipart.chunk_size
    res = client.put(
        multipart_url + "&partNumber={0}".format(1),
        input_stream=BytesIO(data),
    )
    assert res.status_code == expected

    if res.status_code == 200:
        assert res.get_etag()[0] == get_md5(data)

        # Assert content
        with open(multipart.file.uri, "rb") as fp:
            fp.seek(multipart.chunk_size)
            content = fp.read(multipart.chunk_size)
        assert content == data
        assert Part.count(multipart) == 1
        assert Part.get_or_none(multipart, 1).checksum == get_md5(data)


def test_put_not_found(
    client, db, bucket, permissions, multipart, multipart_url, get_md5
):
    """Test invalid multipart id."""
    res = client.put(
        url_for(
            "invenio_files_rest.object_api",
            bucket_id=str(multipart.bucket_id),
            key=multipart.key,
            uploadId="deadbeef-65bd-4d9b-93e2-ec88cc59aec5",
            partNumber=1,
        ),
        input_stream=BytesIO(b"a" * multipart.chunk_size),
    )
    assert res.status_code == 404


def test_put_wrong_sizes(client, db, bucket, multipart, multipart_url, admin_user):
    """Test invalid part sizes."""
    login_user(client, admin_user)

    cases = [
        b"a" * (multipart.chunk_size + 1),
        b"a" * (multipart.chunk_size - 1),
        b"",
    ]
    for data in cases:
        res = client.put(
            multipart_url + "&partNumber={0}".format(1),
            input_stream=BytesIO(data),
        )
        assert res.status_code == 400


def test_put_ngfileupload(client, db, bucket, multipart, multipart_url, admin_user):
    """Test invalid part sizes."""
    login_user(client, admin_user)

    res = client.put(
        multipart_url,
        data={
            "_chunkNumber": "0",
            "_currentChunkSize": str(multipart.chunk_size),
            "_chunkSize": str(multipart.chunk_size),
            "_totalSize": str(multipart.size),
            "file": (BytesIO(b"a" * multipart.chunk_size), multipart.key),
        },
    )
    assert res.status_code == 200


def test_put_invalid_part_number(
    client, db, bucket, multipart, multipart_url, admin_user
):
    """Test invalid part number."""
    login_user(client, admin_user)

    data = b"a" * multipart.chunk_size
    for c in [400, 2000, "a"]:
        res = client.put(
            multipart_url + "&partNumber={0}".format(c),
            input_stream=BytesIO(data),
        )
        assert res.status_code == 400


def test_put_completed_multipart(
    client, db, bucket, multipart, multipart_url, admin_user
):
    """Test uploading to a completed multipart upload."""
    login_user(client, admin_user)

    multipart.completed = True
    db.session.commit()

    res = client.put(
        multipart_url + "&partNumber={0}".format(1),
        input_stream=BytesIO(b"a" * multipart.chunk_size),
    )
    assert res.status_code == 403


def test_put_badstream(
    client, db, bucket, multipart, multipart_url, get_json, admin_user
):
    """Test uploading to a completed multipart upload."""
    login_user(client, admin_user)

    client.put(
        multipart_url + "&partNumber={0}".format(1),
        input_stream=BytesIO(b"a" * multipart.chunk_size),
    )

    # Part exists
    data = get_json(client.get(multipart_url), code=200)
    assert len(data["parts"]) == 1

    pytest.raises(
        ValueError,
        client.put,
        multipart_url + "&partNumber={0}".format(1),
        input_stream=BadBytesIO(b"b" * multipart.chunk_size),
    )

    # Part was removed due to faulty upload which might have written partial
    # content to the file.
    data = get_json(client.get(multipart_url), code=200)
    assert len(data["parts"]) == 0


@pytest.mark.parametrize(
    "user, expected",
    [
        (None, 404),
        ("auth", 404),
        ("objects", 404),
        ("bucket", 200),
        ("location", 200),
    ],
)
def test_get(
    client, db, bucket, permissions, multipart, multipart_url, get_json, user, expected
):
    """Test get parts."""
    Part.create(multipart, 0)
    Part.create(multipart, 1)
    Part.create(multipart, 3)
    db.session.commit()

    login_user(client, permissions[user])

    res = client.get(multipart_url)
    assert res.status_code == expected

    if res.status_code == 200:
        data = get_json(res)
        assert len(data["parts"]) == 3


def test_get_empty(client, multipart, multipart_url, get_json, admin_user):
    """Test get parts when empty."""
    login_user(client, admin_user)

    data = get_json(client.get(multipart_url), code=200)
    assert len(data["parts"]) == 0
    assert data["id"] == str(multipart.upload_id)


def test_get_serialization(client, multipart, multipart_url, get_json, admin_user):
    """Test get parts when empty."""
    login_user(client, admin_user)

    client.put(
        multipart_url + "&partNumber={0}".format(1),
        input_stream=BytesIO(b"a" * multipart.chunk_size),
    )

    data = get_json(client.get(multipart_url), code=200)
    assert len(data["parts"]) == 1

    expected_keys = [
        "part_number",
        "start_byte",
        "end_byte",
        "checksum",
        "created",
        "updated",
    ]
    for k in expected_keys:
        assert k in data["parts"][0]

    expected_keys = [
        "id",
        "bucket",
        "key",
        "completed",
        "size",
        "part_size",
        "links",
        "last_part_number",
        "last_part_size",
        "created",
        "updated",
    ]
    for k in expected_keys:
        assert k in data


@pytest.mark.parametrize(
    "user, expected",
    [
        (None, 404),
        ("auth", 404),
        ("objects", 404),
        ("bucket", 200),
        ("location", 200),
    ],
)
def test_post_complete(
    client,
    headers,
    permissions,
    bucket,
    multipart,
    multipart_url,
    parts,
    get_json,
    user,
    expected,
):
    """Test complete multipart upload."""
    login_user(client, permissions[user])

    # Mock celery task to emulate real usage.
    def _mock_celery_result():
        yield False
        yield False
        merge_multipartobject(str(multipart.upload_id))
        yield True

    result_iter = _mock_celery_result()

    task_result = MagicMock()
    task_result.ready = MagicMock(side_effect=lambda *args: next(result_iter))
    task_result.successful = MagicMock(return_value=True)

    # Complete multipart upload
    with patch("invenio_files_rest.views.merge_multipartobject") as task:
        task.delay = MagicMock(return_value=task_result)

        res = client.post(multipart_url)
        assert res.status_code == expected

        if res.status_code == 200:
            data = get_json(res)
            assert data["completed"] is True
            assert task.called_with(str(multipart.upload_id))
            # Two whitespaces expected to have been sent to client before
            # JSON was sent.
            assert res.data.startswith(b"  {")

            # Multipart object no longer exists
            assert client.get(multipart_url).status_code == 404

            # Object exists
            assert client.get(data["links"]["object"]).status_code == 200


def test_post_complete_fail(
    client, headers, bucket, multipart, multipart_url, parts, get_json, admin_user
):
    """Test completing multipart when merge fails."""
    login_user(client, admin_user)

    # Mock celery task to emulate real usage.
    task_result = MagicMock()
    task_result.ready = MagicMock(side_effect=[False, False, True])
    task_result.successful = MagicMock(return_value=False)

    # Complete multipart upload
    with patch("invenio_files_rest.views.merge_multipartobject") as task:
        task.delay = MagicMock(return_value=task_result)

        res = client.post(multipart_url)
        data = get_json(res, code=200)
        assert res.data.startswith(b"  {")
        assert data["status"] == 500
        assert data["message"] == "Job failed."

        # Multipart object still exists.
        data = get_json(client.get(multipart_url), code=200)
        assert data["completed"] is True

        # Object doesn't exists yet.
        assert client.get(data["links"]["object"]).status_code == 404


def test_post_complete_timeout(
    app, client, headers, bucket, multipart, multipart_url, parts, get_json, admin_user
):
    """Test completing multipart when merge fails."""
    login_user(client, admin_user)

    max_rounds = int(
        app.config["FILES_REST_TASK_WAIT_MAX_SECONDS"]
        // app.config["FILES_REST_TASK_WAIT_INTERVAL"]
    )

    # Mock celery task to emulate real usage.
    task_result = MagicMock()
    task_result.ready = MagicMock(return_value=False)

    # Complete multipart upload
    with patch("invenio_files_rest.views.merge_multipartobject") as task:
        task.delay = MagicMock(return_value=task_result)

        res = client.post(multipart_url)
        data = get_json(res, code=200)
        assert res.data.startswith(b" " * max_rounds)
        assert data["status"] == 500
        assert data["message"] == "Job timed out."

        # Multipart object still exists.
        data = get_json(client.get(multipart_url), code=200)
        assert data["completed"] is True

        # Object doesn't exists yet.
        assert client.get(data["links"]["object"]).status_code == 404


@pytest.mark.parametrize(
    "user, expected",
    [
        (None, 404),
        ("auth", 404),
        ("objects", 404),
        ("bucket", 404),
        ("location", 204),
    ],
)
def test_delete(
    client,
    db,
    bucket,
    multipart,
    multipart_url,
    permissions,
    parts,
    get_json,
    user,
    expected,
):
    """Test complete when parts are missing."""
    assert bucket.size == multipart.size

    login_user(client, permissions[user])
    res = client.delete(multipart_url)
    assert res.status_code == expected

    if res.status_code == 204:
        assert client.get(multipart_url).status_code == 404
        assert MultipartObject.query.count() == 0
        assert Part.query.count() == 0
        assert Bucket.get(bucket.id).size == 0


def test_delete_invalid(client, db, multipart, multipart_url, parts, get_json):
    """Test complete when parts are missing."""
    multipart.complete()
    db.session.commit()
    res = client.delete(multipart_url)
    assert res.status_code == 404


def test_delete_init_not_allowed(client, bucket):
    """Test get request to ?uploads."""
    res = client.delete(
        obj_url(bucket),
        query_string="uploads",
    )
    assert res.status_code == 405


@pytest.mark.parametrize(
    "user, expected",
    [
        (None, 404),
        ("auth", 404),
        ("objects", 404),
        ("bucket", 404),
        ("location", 200),
    ],
)
def test_get_listuploads(
    client,
    db,
    bucket,
    multipart,
    multipart_url,
    permissions,
    parts,
    get_json,
    user,
    expected,
):
    """Test get list of multipart uploads in bucket."""
    login_user(client, permissions[user])

    res = client.get(
        url_for(
            "invenio_files_rest.bucket_api",
            bucket_id=str(bucket.id),
        )
        + "?uploads"
    )
    assert res.status_code == expected


def test_already_exhausted_input_stream(app, client, db, bucket, admin_user):
    """Test server error when file stream is already read."""
    key = "test.json"
    data = b'{"json": "file"}'
    object_url = url_for("invenio_files_rest.object_api", bucket_id=bucket.id, key=key)
    # Add a new before request hook which reads the incoming request payload.
    # This simulates what happens when Sentry's raven-python library when it
    # reads the JSON payloads, breaking the upload of JSON files
    # (`application/json`).

    def consume_request_input_stream(*args):
        """Reads input stream object."""
        from flask import request

        request.data

    app.before_request(consume_request_input_stream)
    login_user(client, admin_user)
    resp = client.put(
        object_url,
        input_stream=BytesIO(data),
    )
    assert resp.status_code == 500
    resp = client.post(
        object_url,
        input_stream=BytesIO(data),
    )
    assert resp.status_code == 500



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Errors for Invenio-Files-REST."""

from invenio_rest.errors import RESTException


class FilesException(RESTException):
    """Base exception for all errors."""

    code = 500


class StorageError(FilesException):
    """Exception raised when a storage operation fails."""

    def get_errors(self):
        """Get errors.

        :returns: A string with the error message.
        """
        if isinstance(self.errors, str):
            return self.errors
        return super(StorageError, self).get_errors()


class UnexpectedFileSizeError(StorageError):
    """Exception raised when a file does not match its expected size."""

    code = 400
    description = "Content-Length does not match file size."


class InvalidOperationError(FilesException):
    """Exception raised when an invalid operation is performed."""

    code = 403


class MissingQueryParameter(FilesException):
    """Exception raised when missing a query parameter."""

    code = 400
    description = "Missing required query argument '{arg_name}'"

    def __init__(self, arg_name, **kwargs):
        """Initialize RESTException."""
        self.arg_name = arg_name
        super(MissingQueryParameter, self).__init__(**kwargs)

    def get_description(self, environ=None):
        """Get the description."""
        return self.description.format(arg_name=self.arg_name)


class FileInstanceAlreadySetError(InvalidOperationError):
    """Exception raised when file instance already set on object."""


class FileInstanceUnreadableError(InvalidOperationError):
    """Exception raised when trying to get an unreadable file."""

    code = 503
    description = "File storage is offline."


class BucketLockedError(InvalidOperationError):
    """Exception raised when a bucket is locked."""

    code = 403
    description = "Bucket is locked for modifications."


class InvalidKeyError(InvalidOperationError):
    """Invalid key."""

    code = 400
    description = "Filename is too long."


class FileSizeError(StorageError):
    """Exception raised when a file larger than allowed."""

    code = 400


class MultipartException(FilesException):
    """Exception for multipart objects."""


class MultipartAlreadyCompleted(MultipartException):
    """Exception raised when multipart object is already completed."""

    code = 403
    description = "Multipart upload is already completed."


class MultipartNotCompleted(MultipartException):
    """Exception raised when multipart object is not already completed."""

    code = 400
    description = "Multipart upload is already completed."


class MultipartInvalidChunkSize(MultipartException):
    """Exception raised when multipart object is already completed."""

    code = 400
    description = "Invalid part size."


class MultipartInvalidPartNumber(MultipartException):
    """Exception raised when multipart object is already completed."""

    code = 400
    description = "Invalid part number."


class MultipartInvalidSize(MultipartException):
    """Exception raised when multipart object is already completed."""

    code = 400
    description = "Invalid file size."


class MultipartMissingParts(MultipartException):
    """Exception raised when multipart object is already completed."""

    code = 400
    description = "Not all parts have been uploaded."


class MultipartNoPart(MultipartException):
    """Exception raised by part factories when no part was detected."""

    code = 400
    description = "No upload part detected in request."


class InvalidTagError(InvalidOperationError):
    """Invalid tag key and/or value."""

    code = 400
    description = "Too long/short tag key or tag value."


class DuplicateTagError(InvalidOperationError):
    """Invalid tag key and/or value."""

    code = 400
    description = "Duplicate tag key"


class ExhaustedStreamError(FilesException):
    """The incoming file stream has been already consumed."""

    code = 500



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Storage module tests."""

import pytest

from invenio_files_rest.helpers import make_path


def test_make_path():
    """Test path for files."""
    myid = "deadbeef-dead-dead-dead-deaddeafbeef"
    base = "/base"
    f = "data"

    assert (
        make_path(base, myid, f, 1, 1)
        == "/base/d/eadbeef-dead-dead-dead-deaddeafbeef/data"
    )
    assert (
        make_path(base, myid, f, 3, 1)
        == "/base/d/e/a/dbeef-dead-dead-dead-deaddeafbeef/data"
    )
    assert (
        make_path(base, myid, f, 1, 3)
        == "/base/dea/dbeef-dead-dead-dead-deaddeafbeef/data"
    )
    assert (
        make_path(base, myid, f, 2, 2)
        == "/base/de/ad/beef-dead-dead-dead-deaddeafbeef/data"
    )

    pytest.raises(AssertionError, make_path, base, myid, f, 1, 50)
    pytest.raises(AssertionError, make_path, base, myid, f, 50, 1)
    pytest.raises(AssertionError, make_path, base, myid, f, 50, 50)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
# Copyright (C) 2022 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Files download/upload REST API similar to S3 for Invenio."""

import uuid
from functools import partial, wraps
from urllib.parse import parse_qsl

from flask import Blueprint, abort, current_app, request
from flask_login import current_user
from invenio_db import db
from invenio_rest import ContentNegotiatedMethodView
from marshmallow import missing
from webargs import fields
from webargs.flaskparser import use_kwargs

from .errors import (
    DuplicateTagError,
    ExhaustedStreamError,
    FileSizeError,
    InvalidTagError,
    MissingQueryParameter,
    MultipartInvalidChunkSize,
)
from .models import Bucket, MultipartObject, ObjectVersion, ObjectVersionTag, Part
from .proxies import current_files_rest, current_permission_factory
from .serializer import json_serializer
from .signals import file_deleted, file_downloaded, file_uploaded
from .tasks import merge_multipartobject, remove_file_data

blueprint = Blueprint(
    "invenio_files_rest",
    __name__,
    url_prefix="/files",
)


#
# Helpers
#
def as_uuid(value):
    """Convert value to UUID."""
    try:
        return uuid.UUID(value)
    except ValueError:
        abort(404)


def minsize_validator(value):
    """Validate Content-Length header.

    :raises invenio_files_rest.errors.FileSizeError: If the value is less
        than :data:`invenio_files_rest.config.FILES_REST_MIN_FILE_SIZE` size.
    """
    if value < current_app.config["FILES_REST_MIN_FILE_SIZE"]:
        raise FileSizeError()


def invalid_subresource_validator(value):
    """Ensure subresource."""
    abort(405)


def validate_tag(key, value):
    """Validate a tag.

    Keys must be less than 128 chars and values must be less than 256 chars.
    """
    # Note, parse_sql does not include a keys if the value is an empty string
    # (e.g. 'key=&test=a'), and thus technically we should not get strings
    # which have zero length.
    klen = len(key)
    vlen = len(value)

    return klen > 0 and klen < 256 and vlen > 0 and vlen < 256


def parse_header_tags():
    """Parse tags specified in the HTTP request header."""
    # Get the value of the custom HTTP header and interpret it as an query
    # string
    qs = request.headers.get(current_app.config["FILES_REST_FILE_TAGS_HEADER"], "")

    tags = {}
    for key, value in parse_qsl(qs):
        # Check for duplicate keys
        if key in tags:
            raise DuplicateTagError()
        # Check for too short/long keys and values.
        if not validate_tag(key, value):
            raise InvalidTagError()
        tags[key] = value
    return tags or None


#
# Part upload factories
#
@use_kwargs(
    {
        "part_number": fields.Int(
            metadata={
                "load_from": "partNumber",
                "location": "query",
            },
            required=True,
            data_key="partNumber",
        ),
        "content_length": fields.Int(
            metadata={
                "load_from": "Content-Length",
                "location": "headers",
            },
            required=True,
            validate=minsize_validator,
            data_key="Content-Length",
        ),
        "content_type": fields.Str(
            metadata={
                "load_from": "Content-Type",
                "location": "headers",
            },
            data_key="Content-Type",
        ),
        "content_md5": fields.Str(
            metadata={
                "load_from": "Content-MD5",
                "location": "headers",
            },
            data_key="Content-MD5",
        ),
    }
)
def default_partfactory(
    part_number=None, content_length=None, content_type=None, content_md5=None
):
    """Get default part factory.

    :param part_number: The part number. (Default: ``None``)
    :param content_length: The content length. (Default: ``None``)
    :param content_type: The HTTP Content-Type. (Default: ``None``)
    :param content_md5: The content MD5. (Default: ``None``)
    :returns: The content length, the part number, the stream, the content
        type, MD5 of the content.
    """
    return content_length, part_number, request.stream, content_type, content_md5, None


@use_kwargs(
    {
        "content_md5": fields.Str(
            metadata={
                "load_from": "Content-MD5",
                "location": "headers",
            },
            data_key="Content-MD5",
            load_default=None,
        ),
        "content_length": fields.Int(
            metadata={
                "load_from": "Content-Length",
                "location": "headers",
            },
            data_key="Content-Length",
            required=True,
            validate=minsize_validator,
        ),
        "content_type": fields.Str(
            metadata={
                "load_from": "Content-Type",
                "location": "headers",
            },
            data_key="Content-Type",
            load_default="",
        ),
    }
)
def stream_uploadfactory(content_md5=None, content_length=None, content_type=None):
    """Get default put factory.

    If Content-Type is ``'multipart/form-data'`` then the stream is aborted.

    :param content_md5: The content MD5. (Default: ``None``)
    :param content_length: The content length. (Default: ``None``)
    :param content_type: The HTTP Content-Type. (Default: ``None``)
    :returns: The stream, content length, MD5 of the content.
    """
    if content_type.startswith("multipart/form-data"):
        abort(422)

    return request.stream, content_length, content_md5, parse_header_tags()


@use_kwargs(
    {
        "part_number": fields.Int(
            metadata={
                "load_from": "_chunkNumber",
                "location": "form",
            },
            data_key="_chunkNumber",
            required=True,
        ),
        "content_length": fields.Int(
            metadata={
                "load_from": "_currentChunkSize",
                "location": "form",
            },
            data_key="_currentChunkSize",
            required=True,
            validate=minsize_validator,
        ),
        "uploaded_file": fields.Raw(
            metadata={
                "load_from": "file",
                "location": "files",
            },
            data_key="file",
            required=True,
        ),
    }
)
def ngfileupload_partfactory(part_number=None, content_length=None, uploaded_file=None):
    """Part factory for ng-file-upload.

    :param part_number: The part number. (Default: ``None``)
    :param content_length: The content length. (Default: ``None``)
    :param uploaded_file: The upload request. (Default: ``None``)
    :returns: The content length, part number, stream, HTTP Content-Type
        header.
    """
    return (
        content_length,
        part_number,
        uploaded_file.stream,
        uploaded_file.headers.get("Content-Type"),
        None,
        None,
    )


@use_kwargs(
    {
        "content_length": fields.Int(
            metadata={
                "load_from": "_totalSize",
                "location": "form",
            },
            data_key="_totalSize",
            required=True,
        ),
        "content_type": fields.Str(
            metadata={
                "load_from": "Content-Type",
                "location": "headers",
            },
            data_key="Content-Type",
            required=True,
        ),
        "uploaded_file": fields.Raw(
            metadata={
                "load_from": "file",
                "location": "files",
            },
            data_key="file",
            required=True,
        ),
    }
)
def ngfileupload_uploadfactory(
    content_length=None, content_type=None, uploaded_file=None
):
    """Get default put factory.

    If Content-Type is ``'multipart/form-data'`` then the stream is aborted.

    :param content_length: The content length. (Default: ``None``)
    :param content_type: The HTTP Content-Type. (Default: ``None``)
    :param uploaded_file: The upload request. (Default: ``None``)
    :param file_tags_header: The file tags. (Default: ``None``)
    :returns: A tuple containing stream, content length, and empty header.
    """
    if not content_type.startswith("multipart/form-data"):
        abort(422)

    return uploaded_file.stream, content_length, None, parse_header_tags()


#
# Object retrieval
#
def pass_bucket(f):
    """Decorate to retrieve a bucket."""

    @wraps(f)
    def decorate(*args, **kwargs):
        bucket_id = kwargs.pop("bucket_id")
        bucket = Bucket.get(as_uuid(bucket_id))
        if not bucket:
            abort(404, "Bucket does not exist.")
        return f(bucket=bucket, *args, **kwargs)

    return decorate


def pass_multipart(with_completed=False):
    """Decorate to retrieve an object."""

    def decorate(f):
        @wraps(f)
        def inner(self, bucket, key, upload_id, *args, **kwargs):
            obj = MultipartObject.get(
                bucket, key, upload_id, with_completed=with_completed
            )
            if obj is None:
                abort(404, "uploadId does not exists.")
            return f(self, obj, *args, **kwargs)

        return inner

    return decorate


def ensure_input_stream_is_not_exhausted(f):
    """Make sure that the input stream has not been read already."""

    @wraps(f)
    def decorate(*args, **kwargs):
        if request.content_length and request.stream.is_exhausted:
            raise ExhaustedStreamError()
        return f(*args, **kwargs)

    return decorate


#
# Permission checking
#
def check_permission(permission, hidden=True):
    """Check if permission is allowed.

    If permission fails then the connection is aborted.

    :param permission: The permission to check.
    :param hidden: Determine if a 404 error (``True``) or 401/403 error
        (``False``) should be returned if the permission is rejected (i.e.
        hide or reveal the existence of a particular object).
    """
    if permission is not None and not permission.can():
        if hidden:
            abort(404)
        else:
            if current_user.is_authenticated:
                abort(403, "You do not have a permission for this action")
            abort(401)


def need_permissions(object_getter, action, hidden=True):
    """Get permission for buckets or abort.

    :param object_getter: The function used to retrieve the object and pass it
        to the permission factory.
    :param action: The action needed.
    :param hidden: Determine which kind of error to return. (Default: ``True``)
    """

    def decorator_builder(f):
        @wraps(f)
        def decorate(*args, **kwargs):
            check_permission(
                current_permission_factory(
                    object_getter(*args, **kwargs),
                    action(*args, **kwargs) if callable(action) else action,
                ),
                hidden=hidden,
            )
            return f(*args, **kwargs)

        return decorate

    return decorator_builder


need_location_permission = partial(
    need_permissions, lambda *args, **kwargs: kwargs.get("location")
)


need_bucket_permission = partial(
    need_permissions, lambda *args, **kwargs: kwargs.get("bucket")
)


#
# REST resources
#
class LocationResource(ContentNegotiatedMethodView):
    """Service resource."""

    def __init__(self, *args, **kwargs):
        """Instantiate content negotiated view."""
        super(LocationResource, self).__init__(*args, **kwargs)

    @need_location_permission("location-update", hidden=False)
    def post(self):
        """Create bucket."""
        with db.session.begin_nested():
            bucket = Bucket.create()
        db.session.commit()
        return self.make_response(
            data=bucket,
            context={
                "class": Bucket,
            },
        )


class BucketResource(ContentNegotiatedMethodView):
    """Bucket item resource."""

    get_args = {
        "versions": fields.Raw(
            metadata={"location": "query"},
        ),
        "uploads": fields.Raw(
            metadata={"location": "query"},
        ),
    }

    def __init__(self, *args, **kwargs):
        """Instantiate content negotiated view."""
        super(BucketResource, self).__init__(*args, **kwargs)

    @need_permissions(lambda self, bucket: bucket, "bucket-listmultiparts")
    def multipart_listuploads(self, bucket):
        """List objects in a bucket.

        :param bucket: A :class:`invenio_files_rest.models.Bucket` instance.
        :returns: The Flask response.
        """
        return self.make_response(
            data=MultipartObject.query_by_bucket(bucket).limit(1000).all(),
            context={
                "class": MultipartObject,
                "bucket": bucket,
                "many": True,
            },
        )

    @need_permissions(
        lambda self, bucket, versions: bucket,
        "bucket-read",
    )
    def listobjects(self, bucket, versions):
        """List objects in a bucket.

        :param bucket: A :class:`invenio_files_rest.models.Bucket` instance.
        :returns: The Flask response.
        """
        if versions is not missing:
            check_permission(
                current_permission_factory(bucket, "bucket-read-versions"), hidden=False
            )
        return self.make_response(
            data=ObjectVersion.get_by_bucket(
                bucket.id, versions=versions is not missing
            )
            .limit(1000)
            .all(),
            context={
                "class": ObjectVersion,
                "bucket": bucket,
                "many": True,
            },
        )

    @use_kwargs(get_args)
    @pass_bucket
    def get(self, bucket=None, versions=missing, uploads=missing):
        """Get list of objects in the bucket.

        :param bucket: A :class:`invenio_files_rest.models.Bucket` instance.
        :returns: The Flask response.
        """
        if uploads is not missing:
            return self.multipart_listuploads(bucket)
        else:
            return self.listobjects(bucket, versions)

    @pass_bucket
    @need_bucket_permission("bucket-read")
    def head(self, bucket=None, **kwargs):
        """Check the existence of the bucket."""


class ObjectResource(ContentNegotiatedMethodView):
    """Object item resource."""

    delete_args = {
        "version_id": fields.UUID(
            metadata={
                "load_from": "versionId",
                "location": "query",
            },
            data_key="versionId",
            load_default=None,
        ),
        "upload_id": fields.UUID(
            metadata={
                "load_from": "uploadId",
                "location": "query",
            },
            data_key="uploadId",
            load_default=None,
        ),
        "uploads": fields.Raw(
            metadata={
                "location": "query",
            },
            validate=invalid_subresource_validator,
        ),
    }

    get_args = dict(
        delete_args,
        download=fields.Raw(
            metadata={
                "location": "query",
            },
            load_default=None,
        ),
    )

    post_args = {
        "uploads": fields.Raw(
            metadata={
                "location": "query",
            }
        ),
        "upload_id": fields.UUID(
            metadata={
                "location": "query",
                "load_from": "uploadId",
            },
            data_key="uploadId",
            load_default=None,
        ),
    }

    put_args = {
        "upload_id": fields.UUID(
            metadata={
                "location": "query",
                "load_from": "uploadId",
            },
            data_key="uploadId",
            load_default=None,
        ),
    }

    multipart_init_args = {
        "size": fields.Int(
            metadata={
                "locations": ("query", "json"),
            },
            load_default=None,
        ),
        "part_size": fields.Int(
            metadata={
                "locations": ("query", "json"),
                "load_from": "partSize",
            },
            load_default=None,
            data_key="partSize",
        ),
    }

    def __init__(self, *args, **kwargs):
        """Instantiate content negotiated view."""
        super(ObjectResource, self).__init__(*args, **kwargs)

    #
    # ObjectVersion helpers
    #
    @staticmethod
    def check_object_permission(obj):
        """Retrieve object and abort if it doesn't exists."""
        check_permission(current_permission_factory(obj, "object-read"))
        if not obj.is_head:
            check_permission(
                current_permission_factory(obj, "object-read-version"), hidden=False
            )

    @classmethod
    def get_object(cls, bucket, key, version_id):
        """Retrieve object and abort if it doesn't exist.

        If the file is not found, the connection is aborted and the 404
        error is returned.

        :param bucket: The bucket (instance or id) to get the object from.
        :param key: The file key.
        :param version_id: The version ID.
        :returns: A :class:`invenio_files_rest.models.ObjectVersion` instance.
        """
        obj = ObjectVersion.get(bucket, key, version_id=version_id)
        if not obj:
            abort(404, "Object does not exists.")

        cls.check_object_permission(obj)

        return obj

    def create_object(self, bucket, key):
        """Create a new object.

        :param bucket: The bucket (instance or id) to get the object from.
        :param key: The file key.
        :returns: A Flask response.
        """
        # Initial validation of size based on Content-Length.
        # User can tamper with Content-Length, so this is just an initial up
        # front check. The storage subsystem must validate the size limit as
        # well.
        stream, content_length, content_md5, tags = current_files_rest.upload_factory()

        size_limit = bucket.size_limit
        if content_length and size_limit and content_length > size_limit:
            desc = (
                "File size limit exceeded."
                if isinstance(size_limit, int)
                else size_limit.reason
            )
            raise FileSizeError(description=desc)

        with db.session.begin_nested():
            obj = ObjectVersion.create(bucket, key)
            obj.set_contents(stream, size=content_length, size_limit=size_limit)
            # Check add tags
            if tags:
                for key, value in tags.items():
                    ObjectVersionTag.create(obj, key, value)

        db.session.commit()
        file_uploaded.send(current_app._get_current_object(), obj=obj)
        return self.make_response(
            data=obj,
            context={
                "class": ObjectVersion,
                "bucket": bucket,
            },
            etag=obj.file.checksum,
        )

    @need_permissions(
        lambda self, bucket, obj, *args: obj,
        "object-delete",
        hidden=False,  # Because get_object permission check has already run
    )
    def delete_object(self, bucket, obj, version_id):
        """Delete an existing object.

        :param bucket: The bucket (instance or id) to get the object from.
        :param obj: A :class:`invenio_files_rest.models.ObjectVersion`
            instance.
        :param version_id: The version ID.
        :returns: A Flask response.
        """
        if version_id is None:
            # Create a delete marker.
            with db.session.begin_nested():
                ObjectVersion.delete(bucket, obj.key)
        else:
            # Permanently delete specific object version.
            check_permission(
                current_permission_factory(bucket, "object-delete-version"),
                hidden=False,
            )
            obj.remove()
            # Set newest object as head
            if obj.is_head:
                latest = ObjectVersion.get_versions(
                    obj.bucket, obj.key, desc=True
                ).first()
                if latest:
                    latest.is_head = True

            if obj.file_id:
                remove_file_data.delay(str(obj.file_id))
        db.session.commit()
        file_deleted.send(current_app._get_current_object(), obj=obj)
        return self.make_response("", 204)

    @staticmethod
    def send_object(
        bucket,
        obj,
        expected_chksum=None,
        logger_data=None,
        restricted=True,
        as_attachment=False,
    ):
        """Send an object for a given bucket.

        :param bucket: The bucket (instance or id) to get the object from.
        :param obj: A :class:`invenio_files_rest.models.ObjectVersion`
            instance.
        :params expected_chksum: Expected checksum.
        :param logger_data: The python logger.
        :param kwargs: Keyword arguments passed to ``Object.send_file()``
        :returns: A Flask response.
        """
        if not obj.is_head:
            check_permission(
                current_permission_factory(obj, "object-read-version"), hidden=False
            )

        if expected_chksum and obj.file.checksum != expected_chksum:
            current_app.logger.warning(
                "File checksum mismatch detected.", extra=logger_data
            )

        file_downloaded.send(current_app._get_current_object(), obj=obj)
        return obj.send_file(restricted=restricted, as_attachment=as_attachment)

    #
    # MultipartObject helpers
    #
    @pass_multipart(with_completed=True)
    @need_permissions(lambda self, multipart: multipart, "multipart-read")
    def multipart_listparts(self, multipart):
        """Get parts of a multipart upload.

        :param multipart: A :class:`invenio_files_rest.models.MultipartObject`
            instance.
        :returns: A Flask response.
        """
        return self.make_response(
            data=Part.query_by_multipart(multipart)
            .order_by(Part.part_number)
            .limit(1000)
            .all(),
            context={
                "class": Part,
                "multipart": multipart,
                "many": True,
            },
        )

    @use_kwargs(multipart_init_args)
    def multipart_init(self, bucket, key, size=None, part_size=None):
        """Initialize a multipart upload.

        :param bucket: The bucket (instance or id) to get the object from.
        :param key: The file key.
        :param size: The total size.
        :param part_size: The part size.
        :raises invenio_files_rest.errors.MissingQueryParameter: If size or
            part_size are not defined.
        :returns: A Flask response.
        """
        if size is None:
            raise MissingQueryParameter("size")
        if part_size is None:
            raise MissingQueryParameter("partSize")
        multipart = MultipartObject.create(bucket, key, size, part_size)
        db.session.commit()
        return self.make_response(
            data=multipart,
            context={
                "class": MultipartObject,
                "bucket": bucket,
            },
        )

    @pass_multipart(with_completed=True)
    def multipart_uploadpart(self, multipart):
        """Upload a part.

        :param multipart: A :class:`invenio_files_rest.models.MultipartObject`
            instance.
        :returns: A Flask response.
        """
        (
            content_length,
            part_number,
            stream,
            content_type,
            content_md5,
            tags,
        ) = current_files_rest.multipart_partfactory()

        if content_length:
            ck = (
                multipart.last_part_size
                if part_number == multipart.last_part_number
                else multipart.chunk_size
            )

            if ck != content_length:
                raise MultipartInvalidChunkSize()

        # Create part
        try:
            p = Part.get_or_create(multipart, part_number)
            p.set_contents(stream)
            db.session.commit()
        except Exception:
            # We remove the Part since incomplete data may have been written to
            # disk (e.g. client closed connection etc.) so it must be
            # reuploaded.
            db.session.rollback()
            Part.delete(multipart, part_number)
            raise
        return self.make_response(
            data=p,
            context={
                "class": Part,
            },
            etag=p.checksum,
        )

    @pass_multipart(with_completed=True)
    def multipart_complete(self, multipart):
        """Complete a multipart upload.

        :param multipart: A :class:`invenio_files_rest.models.MultipartObject`
            instance.
        :returns: A Flask response.
        """
        multipart.complete()
        db.session.commit()

        version_id = str(uuid.uuid4())

        return self.make_response(
            data=multipart,
            context={
                "class": MultipartObject,
                "bucket": multipart.bucket,
                "object_version_id": version_id,
            },
            # This will wait for the result, and send whitespace on the
            # connection until the task has finished (or max timeout reached).
            task_result=merge_multipartobject.delay(
                str(multipart.upload_id),
                version_id=version_id,
            ),
        )

    @pass_multipart()
    @need_permissions(
        lambda self, multipart: multipart,
        "multipart-delete",
    )
    def multipart_delete(self, multipart):
        """Abort a multipart upload.

        :param multipart: A :class:`invenio_files_rest.models.MultipartObject`
            instance.
        :returns: A Flask response.
        """
        multipart.delete()
        db.session.commit()
        if multipart.file_id:
            remove_file_data.delay(str(multipart.file_id))
        return self.make_response("", 204)

    #
    # HTTP methods implementations
    #
    @use_kwargs(get_args)
    @pass_bucket
    def get(
        self,
        bucket=None,
        key=None,
        version_id=None,
        upload_id=None,
        uploads=None,
        download=None,
    ):
        """Get object or list parts of a multipart upload.

        :param bucket: The bucket (instance or id) to get the object from.
            (Default: ``None``)
        :param key: The file key. (Default: ``None``)
        :param version_id: The version ID. (Default: ``None``)
        :param upload_id: The upload ID. (Default: ``None``)
        :param download: The download flag. (Default: ``None``)
        :returns: A Flask response.
        """
        if upload_id:
            return self.multipart_listparts(bucket, key, upload_id)
        else:
            obj = self.get_object(bucket, key, version_id)
            if current_app.config["FILES_REST_XSENDFILE_ENABLED"]:
                response_constructor = current_app.config[
                    "FILES_REST_XSENDFILE_RESPONSE_FUNC"
                ]
                return response_constructor(obj)
            # If 'download' is missing from query string it will have
            # the value None.
            return self.send_object(bucket, obj, as_attachment=download is not None)

    @use_kwargs(post_args)
    @pass_bucket
    @need_bucket_permission("bucket-update")
    @ensure_input_stream_is_not_exhausted
    def post(self, bucket=None, key=None, uploads=missing, upload_id=None):
        """Upload a new object or start/complete a multipart upload.

        :param bucket: The bucket (instance or id) to get the object from.
            (Default: ``None``)
        :param key: The file key. (Default: ``None``)
        :param upload_id: The upload ID. (Default: ``None``)
        :returns: A Flask response.
        """
        if uploads is not missing:
            return self.multipart_init(bucket, key)
        elif upload_id is not None:
            return self.multipart_complete(bucket, key, upload_id)
        abort(403)

    @use_kwargs(put_args)
    @pass_bucket
    @need_bucket_permission("bucket-update")
    @ensure_input_stream_is_not_exhausted
    def put(self, bucket=None, key=None, upload_id=None):
        """Update a new object or upload a part of a multipart upload.

        :param bucket: The bucket (instance or id) to get the object from.
            (Default: ``None``)
        :param key: The file key. (Default: ``None``)
        :param upload_id: The upload ID. (Default: ``None``)
        :returns: A Flask response.
        """
        if upload_id is not None:
            return self.multipart_uploadpart(bucket, key, upload_id)
        else:
            return self.create_object(bucket, key)

    @use_kwargs(delete_args)
    @pass_bucket
    def delete(
        self, bucket=None, key=None, version_id=None, upload_id=None, uploads=None
    ):
        """Delete an object or abort a multipart upload.

        :param bucket: The bucket (instance or id) to get the object from.
            (Default: ``None``)
        :param key: The file key. (Default: ``None``)
        :param version_id: The version ID. (Default: ``None``)
        :param upload_id: The upload ID. (Default: ``None``)
        :returns: A Flask response.
        """
        if upload_id is not None:
            return self.multipart_delete(bucket, key, upload_id)
        else:
            obj = self.get_object(bucket, key, version_id)
            return self.delete_object(bucket, obj, version_id)


#
# Blueprint definition
#
location_view = LocationResource.as_view(
    "location_api",
    serializers={
        "application/json": json_serializer,
    },
)
bucket_view = BucketResource.as_view(
    "bucket_api",
    serializers={
        "application/json": json_serializer,
    },
)
object_view = ObjectResource.as_view(
    "object_api",
    serializers={
        "application/json": json_serializer,
    },
)

blueprint.add_url_rule(
    "",
    view_func=location_view,
)
blueprint.add_url_rule(
    "/<string:bucket_id>",
    view_func=bucket_view,
)
blueprint.add_url_rule(
    "/<string:bucket_id>/<path:key>",
    view_func=object_view,
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Storage module tests."""

import errno
import os
from io import BytesIO
from os.path import dirname, exists, getsize, join
from unittest.mock import patch

import pytest
from fs.errors import DirectoryNotEmpty, FSError

from invenio_files_rest.errors import (
    FileSizeError,
    StorageError,
    UnexpectedFileSizeError,
)
from invenio_files_rest.limiters import FileSizeLimit
from invenio_files_rest.storage import FileStorage, PyFSFileStorage


def test_storage_interface():
    """Test storage interface."""
    s = FileStorage()
    pytest.raises(NotImplementedError, s.open)
    pytest.raises(NotImplementedError, s.initialize)
    pytest.raises(NotImplementedError, s.delete)
    pytest.raises(NotImplementedError, s.save, None)
    pytest.raises(NotImplementedError, s.update, None)
    pytest.raises(NotImplementedError, s.checksum)


def test_pyfs_initialize(pyfs, pyfs_testpath):
    """Test init of files."""
    # Create file object.
    assert not exists(pyfs_testpath)
    uri, size, checksum = pyfs.initialize(size=100)

    assert size == 100
    assert checksum is None
    assert os.stat(pyfs_testpath).st_size == size

    uri, size, checksum = pyfs.initialize()
    assert size == 0
    assert size == os.stat(pyfs_testpath).st_size


def test_pyfs_delete(app, db, dummy_location):
    """Test init of files."""
    testurl = join(dummy_location.uri, "subpath/data")
    s = PyFSFileStorage(testurl)
    s.initialize(size=100)
    assert exists(testurl)

    s.delete()
    assert not exists(testurl)

    s = PyFSFileStorage(join(dummy_location.uri, "anotherpath/data"))

    # this used to raise a ResourceNotFound, but fs>=2.0 raises CreateFailed
    # FSError is their last common ancestor
    pytest.raises(FSError, s.delete)


def test_pyfs_delete_fail(pyfs, pyfs_testpath):
    """Test init of files."""
    pyfs.save(BytesIO(b"somedata"))
    os.rename(pyfs_testpath, join(dirname(pyfs_testpath), "newname"))
    pytest.raises(DirectoryNotEmpty, pyfs.delete)


def test_pyfs_save(pyfs, pyfs_testpath, get_md5):
    """Test basic save operation."""
    data = b"somedata"
    uri, size, checksum = pyfs.save(BytesIO(data))

    assert uri == pyfs_testpath
    assert size == len(data)
    assert checksum == get_md5(data)
    assert exists(pyfs_testpath)
    assert open(pyfs_testpath, "rb").read() == data


def test_pyfs_save_failcleanup(pyfs, pyfs_testpath, get_md5):
    """Test basic save operation."""
    data = b"somedata"

    def fail_callback(total, size):
        assert exists(pyfs_testpath)
        raise Exception("Something bad happened")

    pytest.raises(
        Exception,
        pyfs.save,
        BytesIO(data),
        chunk_size=4,
        progress_callback=fail_callback,
    )
    assert not exists(pyfs_testpath)
    assert not exists(dirname(pyfs_testpath))


def test_pyfs_save_callback(pyfs):
    """Test progress callback."""
    data = b"somedata"

    counter = dict(size=0)

    def callback(total, size):
        counter["size"] = size

    uri, size, checksum = pyfs.save(BytesIO(data), progress_callback=callback)

    assert counter["size"] == len(data)


def test_pyfs_save_limits(pyfs):
    """Test progress callback."""
    data = b"somedata"
    uri, size, checksum = pyfs.save(BytesIO(data), size=len(data))
    assert size == len(data)

    uri, size, checksum = pyfs.save(BytesIO(data), size_limit=len(data))
    assert size == len(data)

    # Size doesn't match
    pytest.raises(UnexpectedFileSizeError, pyfs.save, BytesIO(data), size=len(data) - 1)
    pytest.raises(UnexpectedFileSizeError, pyfs.save, BytesIO(data), size=len(data) + 1)

    # Exceeds size limits
    pytest.raises(
        FileSizeError,
        pyfs.save,
        BytesIO(data),
        size_limit=FileSizeLimit(len(data) - 1, "bla"),
    )


def test_pyfs_update(pyfs, pyfs_testpath, get_md5):
    """Test update of file."""
    pyfs.initialize(size=100)
    pyfs.update(BytesIO(b"cd"), seek=2, size=2)
    pyfs.update(BytesIO(b"ab"), seek=0, size=2)

    with open(pyfs_testpath) as fp:
        content = fp.read()
    assert content[0:4] == "abcd"
    assert len(content) == 100

    # Assert return parameters from update.
    size, checksum = pyfs.update(BytesIO(b"ef"), seek=4, size=2)
    assert size == 2
    assert get_md5(b"ef") == checksum


def test_pyfs_update_fail(pyfs, pyfs_testpath, get_md5):
    """Test update of file."""

    def fail_callback(total, size):
        assert exists(pyfs_testpath)
        raise Exception("Something bad happened")

    pyfs.initialize(size=100)
    pyfs.update(BytesIO(b"ab"), seek=0, size=2)
    pytest.raises(
        Exception,
        pyfs.update,
        BytesIO(b"cdef"),
        seek=2,
        size=4,
        chunk_size=2,
        progress_callback=fail_callback,
    )

    # Partial file can be written to disk!
    with open(pyfs_testpath) as fp:
        content = fp.read()
    assert content[0:4] == "abcd"
    assert content[4:6] != "ef"


def test_pyfs_checksum(get_md5):
    """Test fixity."""
    # Compute checksum of license file/
    with open("LICENSE", "rb") as fp:
        data = fp.read()
        checksum = get_md5(data)

    counter = dict(size=0)

    def callback(total, size):
        counter["size"] = size

    # Now do it with storage interface
    s = PyFSFileStorage("LICENSE", size=getsize("LICENSE"))
    assert checksum == s.checksum(chunk_size=2, progress_callback=callback)
    assert counter["size"] == getsize("LICENSE")

    # No size provided, means progress callback isn't called
    counter["size"] = 0
    s = PyFSFileStorage("LICENSE")
    assert checksum == s.checksum(chunk_size=2, progress_callback=callback)
    assert counter["size"] == 0


def test_pyfs_checksum_fail():
    """Test fixity problems."""

    # Raise an error during checksum calculation
    def callback(total, size):
        raise OSError(errno.EPERM, "Permission")

    s = PyFSFileStorage("LICENSE", size=getsize("LICENSE"))

    pytest.raises(StorageError, s.checksum, progress_callback=callback)


def test_pyfs_send_file(app, pyfs):
    """Test send file."""
    data = b"sendthis"
    uri, size, checksum = pyfs.save(BytesIO(data))

    with app.test_request_context():
        res = pyfs.send_file("myfilename.txt", mimetype="text/plain", checksum=checksum)
        assert res.status_code == 200
        h = res.headers
        assert h["Content-Type"] == "text/plain; charset=utf-8"
        assert h["Content-Length"] == str(size)
        assert h["Content-MD5"] == checksum[4:]
        assert h["ETag"] == '"{0}"'.format(checksum)

        # Content-Type: application/octet-stream
        # ETag: "b234ee4d69f5fce4486a80fdaf4a4263"
        # Last-Modified: Sat, 23 Jan 2016 06:21:04 GMT
        # Cache-Control: max-age=43200, public
        # Expires: Sat, 23 Jan 2016 19:21:04 GMT
        # Date: Sat, 23 Jan 2016 07:21:04 GMT

        res = pyfs.send_file(
            "myfilename.txt", mimetype="text/plain", checksum="crc32:test"
        )
        assert res.status_code == 200
        assert "Content-MD5" not in dict(res.headers)

        # Test for absence of Content-Disposition header to make sure that
        # it's not present when as_attachment=False
        res = pyfs.send_file(
            "myfilename.txt",
            mimetype="text/plain",
            checksum=checksum,
            as_attachment=False,
        )
        assert res.status_code == 200
        assert "attachment" not in res.headers["Content-Disposition"]


def test_pyfs_send_file_for_download(app, pyfs):
    """Test send file."""
    data = b"sendthis"
    uri, size, checksum = pyfs.save(BytesIO(data))

    with app.test_request_context():
        # Test for presence of Content-Disposition header to make sure that
        # it's present when as_attachment=True
        res = pyfs.send_file(
            "myfilename.txt",
            mimetype="text/plain",
            checksum=checksum,
            as_attachment=True,
        )
        assert res.status_code == 200
        assert (
            res.headers["Content-Disposition"] == "attachment; filename=myfilename.txt"
        )


def test_pyfs_send_file_xss_prevention(app, pyfs):
    """Test send file."""
    data = b'<html><body><script>alert("xss");</script></body></html>'
    uri, size, checksum = pyfs.save(BytesIO(data))

    with app.test_request_context():
        res = pyfs.send_file("myfilename.html", mimetype="text/html", checksum=checksum)
        assert res.status_code == 200
        h = res.headers
        assert h["Content-Type"] == "text/plain; charset=utf-8"
        assert h["Content-Length"] == str(size)
        assert h["Content-MD5"] == checksum[4:]
        assert h["ETag"] == '"{0}"'.format(checksum)
        # XSS prevention
        assert h["Content-Security-Policy"] == "default-src 'none';"
        assert h["X-Content-Type-Options"] == "nosniff"
        assert h["X-Download-Options"] == "noopen"
        assert h["X-Permitted-Cross-Domain-Policies"] == "none"
        assert h["X-Frame-Options"] == "deny"
        assert h["X-XSS-Protection"] == "1; mode=block"
        assert h["Content-Disposition"] == "inline"

        # Image
        h = pyfs.send_file("image.png", mimetype="image/png").headers
        assert h["Content-Type"] == "image/png"
        assert h["Content-Disposition"] == "inline"

        # README text file
        h = pyfs.send_file("README").headers
        assert h["Content-Type"] == "text/plain; charset=utf-8"
        assert h["Content-Disposition"] == "inline"

        # Zip
        h = pyfs.send_file("archive.zip").headers
        assert h["Content-Type"] == "application/octet-stream"
        assert h["Content-Disposition"] == "attachment; filename=archive.zip"

        # PDF
        h = pyfs.send_file("doc.pdf").headers
        assert h["Content-Type"] == "application/octet-stream"
        assert h["Content-Disposition"] == "attachment; filename=doc.pdf"


def test_pyfs_send_file_fail(app, pyfs):
    """Test send file."""
    pyfs.save(BytesIO(b"content"))

    with patch("invenio_files_rest.storage.base.send_stream") as send_stream:
        send_stream.side_effect = OSError(errno.EPERM, "Permission problem")
        with app.test_request_context():
            pytest.raises(StorageError, pyfs.send_file, "test.txt")


def test_pyfs_copy(pyfs, dummy_location):
    """Test send file."""
    s = PyFSFileStorage(join(dummy_location.uri, "anotherpath/data"))
    s.save(BytesIO(b"otherdata"))

    pyfs.copy(s)
    fp = pyfs.open()
    assert fp.read() == b"otherdata"


def test_non_unicode_filename(app, pyfs):
    """Test sending the non-unicode filename in the header."""
    data = b"HelloWorld"
    uri, size, checksum = pyfs.save(BytesIO(data))

    with app.test_request_context():
        res = pyfs.send_file(
            "żółć.dat", mimetype="application/octet-stream", checksum=checksum
        )
        assert res.status_code == 200
        assert set(res.headers["Content-Disposition"].split("; ")) == set(
            [
                "attachment",
                "filename=zoc.dat",
                "filename*=UTF-8''%C5%BC%C3%B3%C5%82%C4%87.dat",
            ]
        )

    with app.test_request_context():
        res = pyfs.send_file("żółć.txt", mimetype="text/plain", checksum=checksum)
        assert res.status_code == 200
        assert res.headers["Content-Disposition"] == "inline"



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test of multipart objects."""

import hashlib
from io import BytesIO
from os.path import exists

from invenio_files_rest.models import Bucket, MultipartObject, ObjectVersion, Part


def make_stream(size):
    """Make a stream of a given size."""
    s = BytesIO()
    s.seek(size - 1)
    s.write(b"\0")
    s.seek(0)
    return s


def test_multipart_creation(app, db, bucket):
    """Test multipart creation."""
    mp = MultipartObject.create(bucket, "test.txt", 100, 20)
    db.session.commit()
    assert mp.upload_id
    assert mp.size == 100
    assert mp.chunk_size == 20
    assert mp.completed is False
    assert mp.bucket.size == 100
    assert exists(mp.file.uri)


def test_multipart_last_part(app, db, bucket):
    """Test multipart creation."""
    mp = MultipartObject.create(bucket, "test.txt", 100, 20)
    assert mp.last_part_size == 0
    assert mp.last_part_number == 4
    mp = MultipartObject.create(bucket, "test.txt", 101, 20)
    assert mp.last_part_size == 1
    assert mp.last_part_number == 5


def test_part_creation(app, db, bucket, get_md5):
    """Test part creation."""
    assert bucket.size == 0
    mp = MultipartObject.create(bucket, "test.txt", 5, 2)
    db.session.commit()
    assert bucket.size == 5

    Part.create(mp, 2, stream=BytesIO(b"p"))
    Part.create(mp, 0, stream=BytesIO(b"p1"))
    Part.create(mp, 1, stream=BytesIO(b"p2"))
    db.session.commit()
    assert bucket.size == 5

    mp.complete()
    db.session.commit()
    assert bucket.size == 5

    # Assert checksum of part.
    m = hashlib.md5()
    m.update(b"p2")
    assert "md5:{0}".format(m.hexdigest()) == Part.get_or_none(mp, 1).checksum

    obj = mp.merge_parts()
    db.session.commit()
    assert bucket.size == 5

    assert MultipartObject.query.count() == 0
    assert Part.query.count() == 0

    assert obj.file.size == 5
    assert obj.file.checksum == get_md5(b"p1p2p")
    assert obj.file.storage().open().read() == b"p1p2p"
    assert obj.file.writable is False
    assert obj.file.readable is True

    assert obj.version_id == ObjectVersion.get(bucket, "test.txt").version_id


def test_multipart_full(app, db, bucket):
    """Test full multipart object."""
    app.config.update(
        dict(
            FILES_REST_MULTIPART_CHUNKSIZE_MIN=5 * 1024 * 1024,
            FILES_REST_MULTIPART_CHUNKSIZE_MAX=5 * 1024 * 1024 * 1024,
        )
    )

    # Initial parameters
    chunks = 20
    chunk_size = 5 * 1024 * 1024  # 5 MiB
    last_chunk = 1024 * 1024  # 1 MiB
    size = (chunks - 1) * chunk_size + last_chunk

    # Initiate
    mp = MultipartObject.create(bucket, "testfile", size=size, chunk_size=chunk_size)
    db.session.commit()

    # Create parts
    for i in range(chunks):
        part_size = chunk_size if i < chunks - 1 else last_chunk
        Part.create(mp, i, stream=make_stream(part_size))
        db.session.commit()

    # Complete
    mp.complete()
    db.session.commit()

    # Merge parts.
    pre_size = mp.bucket.size
    mp.merge_parts()
    db.session.commit()

    # Test size update
    bucket = Bucket.get(bucket.id)
    assert bucket.size == pre_size

    app.config.update(
        dict(
            FILES_REST_MULTIPART_CHUNKSIZE_MIN=2,
            FILES_REST_MULTIPART_CHUNKSIZE_MAX=20,
        )
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Permissions for files using Invenio-Access."""

from invenio_access import Permission, action_factory

from .models import Bucket, MultipartObject, ObjectVersion

#
# Action needs
#

LocationUpdate = action_factory("files-rest-location-update", parameter=True)
"""Action needed: location update."""

BucketRead = action_factory("files-rest-bucket-read", parameter=True)
"""Action needed: list objects in bucket."""

BucketReadVersions = action_factory("files-rest-bucket-read-versions", parameter=True)
"""Action needed: list object versions in bucket."""

BucketUpdate = action_factory("files-rest-bucket-update", parameter=True)
"""Action needed: create objects and multipart uploads in bucket."""

BucketListMultiparts = action_factory(
    "files-rest-bucket-listmultiparts", parameter=True
)
"""Action needed: list multipart uploads in bucket."""

ObjectRead = action_factory("files-rest-object-read", parameter=True)
"""Action needed: get object in bucket."""

ObjectReadVersion = action_factory("files-rest-object-read-version", parameter=True)
"""Action needed: get object version in bucket."""

ObjectDelete = action_factory("files-rest-object-delete", parameter=True)
"""Action needed: delete object in bucket."""

ObjectDeleteVersion = action_factory("files-rest-object-delete-version", parameter=True)
"""Action needed: permanently delete specific object version in bucket."""


MultipartRead = action_factory("files-rest-multipart-read", parameter=True)
"""Action needed: list parts of a multipart upload in a bucket."""

MultipartDelete = action_factory("files-rest-multipart-delete", parameter=True)
"""Action needed: abort a multipart upload."""


#
# Global action needs
#

location_update_all = LocationUpdate(None)
"""Action needed: update all locations."""

bucket_read_all = BucketRead(None)
"""Action needed: read all buckets."""

bucket_read_versions_all = BucketReadVersions(None)
"""Action needed: read all buckets versions."""

bucket_update_all = BucketUpdate(None)
"""Action needed: update all buckets"""

bucket_listmultiparts_all = BucketListMultiparts(None)
"""Action needed: list all buckets multiparts."""

object_read_all = ObjectRead(None)
"""Action needed: read all objects."""

object_read_version_all = ObjectReadVersion(None)
"""Action needed: read all objects versions."""

object_delete_all = ObjectDelete(None)
"""Action needed: delete all objects."""

object_delete_version_all = ObjectDeleteVersion(None)
"""Action needed: delete all objects versions."""

multipart_read_all = MultipartRead(None)
"""Action needed: read all multiparts."""

multipart_delete_all = MultipartDelete(None)
"""Action needed: delete all multiparts."""


_action2need_map = {
    "location-update": LocationUpdate,
    "bucket-read": BucketRead,
    "bucket-read-versions": BucketReadVersions,
    "bucket-update": BucketUpdate,
    "bucket-listmultiparts": BucketListMultiparts,
    "object-read": ObjectRead,
    "object-read-version": ObjectReadVersion,
    "object-delete": ObjectDelete,
    "object-delete-version": ObjectDeleteVersion,
    "multipart-read": MultipartRead,
    "multipart-delete": MultipartDelete,
}
"""Mapping of action names to action needs."""


def permission_factory(obj, action):
    """Get default permission factory.

    :param obj: An instance of :class:`invenio_files_rest.models.Bucket` or
        :class:`invenio_files_rest.models.ObjectVersion` or
        :class:`invenio_files_rest.models.MultipartObject` or ``None`` if
        the action is global.
    :param action: The required action.
    :raises RuntimeError: If the object is unknown.
    :returns: A :class:`invenio_access.permissions.Permission` instance.
    """
    need_class = _action2need_map[action]

    if obj is None:
        return Permission(need_class(None))

    arg = None
    if isinstance(obj, Bucket):
        arg = str(obj.id)
    elif isinstance(obj, ObjectVersion):
        arg = str(obj.bucket_id)
    elif isinstance(obj, MultipartObject):
        arg = str(obj.bucket_id)
    else:
        raise RuntimeError("Unknown object")

    return Permission(need_class(arg))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create files_objecttags table."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = "8ae99b034410"
down_revision = "f741aa746a7d"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    # table ObjectVersion: modify primary_key
    if op.get_context().dialect.name == "mysql":
        Fk = "fk_files_object_bucket_id_files_bucket"
        op.execute(
            "ALTER TABLE files_object "
            "DROP FOREIGN KEY {0}, DROP PRIMARY KEY, "
            "ADD PRIMARY KEY(version_id), "
            "ADD FOREIGN KEY(bucket_id) "
            "REFERENCES files_bucket(id)".format(Fk)
        )
    else:
        op.drop_constraint("pk_files_object", "files_object", type_="primary")
        op.create_primary_key("pk_files_object", "files_object", ["version_id"])
    op.create_unique_constraint(
        "uq_files_object_bucket_id", "files_object", ["bucket_id", "version_id", "key"]
    )
    # table ObjectVersionTag: create
    op.create_table(
        "files_objecttags",
        sa.Column("version_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column(
            "key", sa.Text().with_variant(mysql.VARCHAR(255), "mysql"), nullable=False
        ),
        sa.Column(
            "value", sa.Text().with_variant(mysql.VARCHAR(255), "mysql"), nullable=False
        ),
        sa.PrimaryKeyConstraint("version_id", "key"),
        sa.ForeignKeyConstraint(
            ["version_id"], ["files_object.version_id"], ondelete="CASCADE"
        ),
    )


def downgrade():
    """Downgrade database."""
    # table ObjectVersionTag
    op.drop_table("files_objecttags")
    # table ObjectVersion: modify primary_key
    if op.get_context().dialect.name == "mysql":
        op.execute(
            "ALTER TABLE files_object "
            "DROP INDEX uq_files_object_bucket_id, "
            "DROP PRIMARY KEY, "
            "ADD PRIMARY KEY(`bucket_id`, `key`, `version_id`)"
        )
    else:
        op.drop_constraint("pk_files_object", "files_object", type_="primary")
        op.create_primary_key(
            "pk_files_object", "files_object", ["bucket_id", "key", "version_id"]
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2024 CERN.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


r"""Invenio-Files-REST module.

This guide will show you how to get started with
Invenio-Files-REST. It assumes that you already have knowledge of
Flask applications and Invenio modules.

It will then explain key topics and concepts of this module.

Getting started
---------------

You will learn how to create a new Location, a Bucket and an ObjectVersion
using the programmatic APIs of Invenio-Files-REST.

First, you will have to setup your virtualenv environment and install
this module along with all it's dependencies.

After that, start a Python shell and execute the following commands:

>>> from flask import Flask
>>> app = Flask('myapp')

This is the initial configuration needed to have things running:

>>> app.config['BROKER_URL'] = 'redis://'
>>> app.config['CELERY_RESULT_BACKEND'] = 'redis://'
>>> app.config['DATADIR'] = 'data'
>>> app.config['FILES_REST_MULTIPART_CHUNKSIZE_MIN'] = 4
>>> app.config['REST_ENABLE_CORS'] = True
>>> app.config['SECRET_KEY'] = 'CHANGEME'
>>> app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite://'
>>> app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

>>> allow_all = lambda *args, **kwargs: \
... type('Allow', (), {'can': lambda self: True})()
>>> app.config['FILES_REST_PERMISSION_FACTORY'] = allow_all

Relevant configuration variables will be explained later on.
Now let's initialize all required Invenio extensions:

>>> import shutil
>>> from os import makedirs
>>> from os.path import dirname, exists, join
>>> from pprint import pprint
>>> import json

>>> from invenio_i18n import Babel, InvenioI18N
>>> from flask_menu import Menu
>>> from invenio_db import InvenioDB, db
>>> from invenio_rest import InvenioREST
>>> from invenio_admin import InvenioAdmin
>>> from invenio_accounts import InvenioAccounts
>>> from invenio_access import InvenioAccess
>>> from invenio_accounts.views import blueprint as accounts_blueprint
>>> from invenio_celery import InvenioCelery
>>> from invenio_files_rest import InvenioFilesREST
>>> from invenio_files_rest.views import blueprint

>>> ext_babel = Babel(app)
>>> ext_menu = Menu(app)
>>> ext_db = InvenioDB(app)
>>> ext_rest = InvenioREST(app)
>>> ext_admin = InvenioAdmin(app)
>>> ext_accounts = InvenioAccounts(app)
>>> ext_access = InvenioAccess(app)
>>> ext_i18n = InvenioI18N(app)

You can now initialize  Invenio-Files-REST. When using Invenio-Files-REST as
dependency of an Invenio applicaton, the REST views are automatically
registered via entry points. For this example, you will have to register
them manually and push a Flask application context:

>>> ext_rest = InvenioFilesREST(app)

>>> app.register_blueprint(accounts_blueprint)
>>> app.register_blueprint(blueprint)

>>> app.app_context().push()

Let's create the database and tables, using an in-memory SQLite database:

>>> db.create_all()

When you setup Invenio-Files-REST for the first time, you will have to define
a default Location. It can be local or remote and it will be accessed via
its URI.

.. _usage-create-location:

Create a location
^^^^^^^^^^^^^^^^^
For this example, you will use a temporary directory:

>>> from invenio_files_rest.models import Location
>>> d = app.config['DATADIR']  # folder `data`
>>> if exists(d): shutil.rmtree(d)
>>> makedirs(d)
>>> loc = Location(name='local', uri=d, default=True)
>>> db.session.add(loc)
>>> db.session.commit()

Create a bucket
^^^^^^^^^^^^^^^
In order to create, modify or delete files, you have to create a files
container first, the Bucket.

>>> from invenio_files_rest.models import Bucket
>>> b1 = Bucket.create(loc)
>>> db.session.commit()

Create objects
^^^^^^^^^^^^^^
Files are represented by ObjectVersions. After creating a bucket, you can now
add files to it, for example:

>>> from io import BytesIO
>>> from invenio_files_rest.models import ObjectVersion
>>> a_file = BytesIO(b"my file contents")
>>> f = ObjectVersion.create(b1, "thesis.pdf", stream=a_file)
>>> db.session.commit()

Retrieve objects
^^^^^^^^^^^^^^^^
You can now retrieve objects. Retrieve the bucket object:

>>> b = Bucket.get(b1.id)

Retrieve all ObjectVersions contained in a bucket:

>>> file_names = [ov.key for ov in ObjectVersion.get_by_bucket(b1.id)]

Retrieve a specific ObjectVersion by filename:

>>> f = ObjectVersion.get(b1.id, "thesis.pdf")

Data model
----------

This is a more in-depth explanation of the concepts introduced in the
:doc:`overview` section.

Buckets
^^^^^^^

A bucket is a container of objects. It is uniquely identified by an ID.
Buckets have a default Location and Storage class.
Individual objects in the bucket can however have different
Locations and Storage classes.

The :code:`size` field stores the current size of the bucket. When a new
object is added or completely removed, its size is updated.

Buckets can have constraints on the maximum amount of objects that they can
contain. It is controlled by the function
:py:func:`invenio_files_rest.limiters.file_size_limiters`:
by default, a new object can be added to the bucket if the maximum
size of the file is lower than
:py:data:`invenio_files_rest.config.FILES_REST_DEFAULT_MAX_FILE_SIZE`
and if the total quota (the sum of sizes of all files) is lower than
:py:data:`invenio_files_rest.config.FILES_REST_DEFAULT_QUOTA_SIZE`.

Buckets can be marked as :code:`locked`. When a bucket is locked, objects
can be retrieved but no object can be added and deleted.

Similarly to objects, bucket can be logically marked as :code:`deleted`
without affecting the actual content. When it is deleted, it simply means
that no objects can be retrieved or added via APIs.

Finally, buckets provide ways to create or synchronize copies: the
:code:`snapshot` operation creates a new copy of a bucket with all
the latest versions of the object it contains, without duplicating files
on disk.
The :code:`sync` operation mirrors objects contained in the source bucket
to the destination bucket.


ObjectVersion
^^^^^^^^^^^^^

ObjectVersions are objects that represent a specific version of a file at
a given point in time. ObjectVersions are uniquely identified by its ID.
They are always contained in an existing Bucket by having the reference
:code:`bucket_id` to it.

An ObjectVersion describes the file (:code:`FileInstance`) that references
with the attribute :code:`file_id`. It also stores some metadata of the file:
the file name, stored in the :code:`key` attribute and the version, stored in
:code:`version_id` attribute. The triplet (bucket_id, key, version_id) is
unique.

For a given :code:`key` in a Bucket, normally the latest version in history
is marked as the :code:`head`.

The :code:`key` has a maximum length defined via
:py:data:`invenio_files_rest.config.FILES_REST_OBJECT_KEY_MAX_LEN`.

ObjectVersion can be marked as deleted by removing its reference to the
file it represents: from the user perspective, deleting a file normally
means adding a new ObjectVersion, which will be the new :code:`head`,
without :code:`file_id`.


FileInstance
^^^^^^^^^^^^

A file instance represents a file on disk. A file instance may be linked
from many objects, while an object can have one and only one file instance.

The file on disk can be retrieved by the file instance :code:`uri`, which
is an absolute path/URI generated when adding the file: the base path is
retrieved from the :code:`Location` used for this file, and the relative
path is assigned by the file's :code:`Storage`. It is responsibility of the
Storage, which is aware of the file system that is managing, to generate
a unique final path for the file. You can modify how the path is generated
with the default storage
:py:func:`invenio_files_rest.storage.pyfs_storage_factory`
by changing
:py:data:`invenio_files_rest.config.FILES_REST_STORAGE_PATH_SPLIT_LENGTH` or
:py:data:`invenio_files_rest.config.FILES_REST_STORAGE_PATH_DIMENSIONS`.

A file instance may not be ready to be accessed, for example in case of
multipart uploads: the attribute :code:`readable` marks it. It can also be
marked as not :code:`writable` if it cannot be deleted or replaced, for
safety reasons.

:code:`checksum`, :code:`last_check_at` and :code:`last_check` are
attributes used to store information about integrity checks.

You can find the documentation of each API in the :doc:`api`.

.. _usage-rest-apis:

REST APIs
---------
REST APIs allow you to perform most of the operations needed when
manipulating files.

By design, Locations cannot be created using REST APIs.
This is because they depend on your physical file storage infrastructure.
You will have to create them in advance when setting up your Invenio
instance.

To be able to run each of the next steps, you can instantiate and start an
Invenio instance as described `here
<https://invenio.readthedocs.io/en/latest/quickstart/quickstart.html#create-an-invenio-instance>`_.

Create a bucket
^^^^^^^^^^^^^^^
A bucket can be created by a POST request to :code:`/files`.
The response will contain the unique ID of the bucket.

.. code-block:: console

   $ curl -X POST http://localhost:5000/api/files

.. code-block:: json

    {
        "max_file_size": null,
        "updated": "2019-05-16T13:07:21.595398+00:00",
        "locked": false,
        "links": {
            "self": "http://localhost:5000/api/files/
                     cb8d0fa7-2349-484b-89cb-16573d57f09e",
            "uploads": "http://localhost:5000/api/files/
                        cb8d0fa7-2349-484b-89cb-16573d57f09e?uploads",
            "versions": "http://localhost:5000/api/files/
                         cb8d0fa7-2349-484b-89cb-16573d57f09e?versions"
        },
        "created": "2019-05-16T13:07:21.595391+00:00",
        "quota_size": null,
        "id": "cb8d0fa7-2349-484b-89cb-16573d57f09e",
        "size": 0
    }


Uploading Files
^^^^^^^^^^^^^^^
You can upload, download and modify single files via REST APIs.
A file is uniquely identified within a bucket by its name and version.
Each file can have multiple versions.

Let's upload a file called :code:`my_file.txt` inside the bucket that
was just created.

.. code-block:: console

   $ BUCKET=cb8d0fa7-2349-484b-89cb-16573d57f09e

   $ echo "my file content" > my_file.txt

   $ curl -i -X PUT --data-binary @my_file.txt \
     "http://localhost:5000/api/files/$BUCKET/my_file.txt"

.. code-block:: json

    {
        "mimetype": "text/plain",
        "updated": "2019-05-16T13:10:22.621533+00:00",
        "links": {
            "self": "http://localhost:5000/api/files/
                     cb8d0fa7-2349-484b-89cb-16573d57f09e/my_file.txt",

            "version": "http://localhost:5000/api/files/
                        cb8d0fa7-2349-484b-89cb-16573d57f09e/my_file.txt?
                        versionId=7f62676d-0b8e-4d77-9687-8465dc506ca8",
            "uploads": "http://localhost:5000/api/files/
                        cb8d0fa7-2349-484b-89cb-16573d57f09e/
                        my_file.txt?uploads"
        },
        "is_head": true,
        "tags": {},
        "checksum": "md5:d7d02c7125bdcdd857eb70cb5f19aecc",
        "created": "2019-05-16T13:10:22.617714+00:00",
        "version_id": "7f62676d-0b8e-4d77-9687-8465dc506ca8",
        "delete_marker": false,
        "key": "my_file.txt",
        "size": 14
    }

If you have a new version of the file, you can upload it to the same bucket
using the same filename. In this case, a new ObjectVersion will be created.

.. code-block:: console

   $ echo "my file content version 2" > my_filev2.txt

   $ curl -i -X PUT --data-binary @my_filev2.txt \
     "http://localhost:5000/api/files/$BUCKET/my_file.txt"

.. code-block:: json

    {
        "mimetype": "text/plain",
        "updated": "2019-05-16T13:11:22.621533+00:00",
        "links": {
            "self": "http://localhost:5000/api/files/
                     cb8d0fa7-2349-484b-89cb-16573d57f09e/my_file.txt",

            "version": "http://localhost:5000/api/files/
                        cb8d0fa7-2349-484b-89cb-16573d57f09e/my_file.txt?
                        versionId=24bf075f-09f4-42f8-9fbe-3f00b8aac3e8",
            "uploads": "http://localhost:5000/api/files/
                        cb8d0fa7-2349-484b-89cb-16573d57f09e/
                        my_file.txt?uploads"
        },
        "is_head": true,
        "tags": {},
        "checksum": "md5:fe76512703258a894e56bac89d2e8dec",
        "created": "2019-05-16T13:11:22.617714+00:00",
        "version_id": "24bf075f-09f4-42f8-9fbe-3f00b8aac3e8",
        "delete_marker": false,
        "key": "my_file.txt",
        "size": 13
    }

When integrating the REST APIs to upload files via a web application, you
might use JavaScript to improve user experience. Invenio-Files-REST provides
out of the box integration with JavaScript uploaders. See the
:ref:`usage-js-uploaders` section for more information.

Invenio-Files-REST also provides different ways to upload large files. See
the :ref:`usage-multipart-upload` and :ref:`usage-large-files` sections
for more information.

Serving files
^^^^^^^^^^^^^

To serve and allow download of files, you can perform a GET request
specifying the bucket and the filename used to upload the file.

.. code-block:: console

   $ curl -i -X GET "http://localhost:5000/api/files/$BUCKET/my_file.txt"

You can also list files or download specific versions of files. See the REST
APIs reference documentation below for more information.

Be aware that there are security implications to take into account when
serving files. See the :ref:`usage-security` for more information.

Invenio-Files-Rest provides also the functionality to serve your files directly
from your external storage. This is achieved by attaching the
`X-Accel-Redirect
<https://www.nginx.com/resources/wiki/start/topics/examples/x-accel/>`_
header to the response, which will then be redirected by
your Web Proxy (e.g. NGINX, Apache) to your external storage, finally streaming
the file directly to the user.
To use this feature you will need to configure your Web Proxy accordingly and
then enable the
:py:data:`invenio_files_rest.config.FILES_REST_XSENDFILE_ENABLED`.

API Reference
^^^^^^^^^^^^^

Default Location
++++++++++++++++

Create a bucket:

.. code-block:: console

    POST /files/

Buckets
+++++++

Check if bucket exists, returning either a 200 or 404:

.. code-block:: console

    HEAD /files/<bucket_id>

Retrieve the latest version of all objects in bucket:

.. code-block:: console

    GET /files/<bucket_id>

Retrieve all versions of files in a bucket:

.. code-block:: console

    GET /files/<bucket_id>?versions

Return list of multipart uploads:

.. code-block:: console

    GET /files/<bucket_id>?uploads

ObjectVersions
++++++++++++++

Initiate multipart upload (see :ref:`usage-multipart-upload`):

.. code-block:: console

    POST /files/<bucket_id>/<file_name>?
         uploads&size=<total_size>&partSize=<part_size>

Finalize multipart upload:

.. code-block:: console

    POST /files/<bucket_id>/<file_name>?uploadId=<upload_id>

Upload a file to a bucket:

.. code-block:: console

    PUT /files/<bucket_id>/<file_name>

Upload part of in-progress multipart upload to a bucket:

.. code-block:: console

    PUT /files/<bucket_id>/<file_name>?uploadId=<upload_id>&part=<part_number>

Retrieve the latest version of a given file. By default, the file is returned
with the header :code:`'Content-Disposition': 'inline'`. Be aware that the
browser will try to preview it.

.. code-block:: console

    GET /files/<bucket_id>/<file_name>

Download the latest version of a given file. It will return the same response
as the request above but with the response header
:code:`'Content-Disposition': 'attachment'` to instruct the browser
trigger a download.

.. code-block:: console

    GET /files/<bucket_id>/<file_name>?download

Retrieve a specific version of a given file:

.. code-block:: console

    GET /files/<bucket_id>/<file_name>?versionId=<version_id>

Retrieve the list of parts of a multipart upload:

.. code-block:: console

    GET /files/<bucket_id>/<file_name>?uploadId=<id_number>

Mark an object as deleted (see :ref:`usage-deleting-files`):

.. code-block:: console

    DELETE /files/<bucket_id>/<file_name>

Permanently erase an object and the physical file on disk:

.. code-block:: console

    DELETE /files/<bucket_id>/<file_name>?versionId=<version_id>

Abort multipart upload:

.. code-block:: console

    DELETE /files/<bucket_id>/<file_name>?uploadId=<upload_id>


.. _usage-deleting-files:

Deleting files
--------------
A delete operation can be of two types:

1. mark an object as deleted, allowing the possibility of restoring
   a deleted file (also called delete marker or soft deletion).
2. permanently remove any trace of an object and referenced file
   on disk (also called hard deletion).

Soft deletion
^^^^^^^^^^^^^
Technically, it creates a new ObjectVersion, that becomes the new :code:`head`,
with no reference to a FileInstance. It is possible to revert it
by getting the previous version.

This operation will not access to the file on disk and it will leave it
untouched.

You can soft delete using REST APIs:

.. code-block:: console

   DELETE /files/<bucket_id>/<file_name>

Hard deletion
^^^^^^^^^^^^^
Given a specific object version, it will delete the ObjectVersion,
the referenced FileInstance and the file on disk. If the deleted version
was the :code:`head`, it will then set the previous object
as the new head.

The deletion of files on disk will not happen immediately. This is because
it is done via an asynchronous task to ensure that the FileInstance is
safely removed from the database in case the low level operation of file
removal on disk fails for any unexpected reason.

You can hard delete a file using REST APIs:

.. code-block:: console

   DELETE /files/<bucket_id>/<file_name>?versionId=<version_id>

REST APIs do not allow to perform delete operations that can affect multiple
objects at the same time. For advanced use cases, you will to use the
Invenio-Files-REST APIs programmatically.

.. note::
    For safety reasons, the deletion will fail if the file that you want
    to delete is referenced by multiple ObjectVersions, for example
    in case of Buckets snapshots.

Authorization
-------------
Invenio-Files-REST relies on `Invenio-Access
<https://invenio-access.readthedocs.io>`_ to implement files authorization.
The following documentation assumes that you already have knowledge of how
authorization works on Invenio.

Invenio-Files-REST defines a set of actions for operations on Bucket and
ObjectVersions that can be used to implement authorization as you need:

    - files-rest-location-update
    - files-rest-bucket-read
    - files-rest-bucket-read-versions
    - files-rest-bucket-update
    - files-rest-bucket-listmultiparts
    - files-rest-object-read
    - files-rest-object-read-version
    - files-rest-object-delete
    - files-rest-object-delete-version
    - files-rest-multipart-read
    - files-rest-multipart-delete

Response codes
^^^^^^^^^^^^^^
If the authorization for an action fails, Invenio-Files-REST normally returns
a :code:`403` response code for authenticated users, :code:`401` otherwise.
For security reasons, when trying to retrieve an unauthorized file, it will
return a :code:`404` instead to hide the existence or
non-existence of the file.

Authorization definition
^^^^^^^^^^^^^^^^^^^^^^^^
The default permission factory
:py:data:`invenio_files_rest.permissions.permission_factory` will authorize
users that has :code:`Needs` that fulfill the actions listed above. This means
that by default no user will be authorized (with the exception of
any :code:`superuser`).

Depending on how you are planning to integrate Invenio-Files-REST in your
Invenio application, you might want to decide how to give permissions for
operations on files.

If you plan to give authorization to specific users or roles, you can use the
default permission factory and assign user or roles to the actions listed
above as described in the Invenio-Access documentation.

If instead you want to define permissions based on other object, for example
on records to which the files are attached to, then you will have to
define your own permission factory and used via the configuration variable
:py:data:`invenio_files_rest.config.FILES_REST_PERMISSION_FACTORY`.

See :mod:`invenio_files_rest.permissions` for more documentation.


.. _usage-security:

Security
--------
When serving files, you will have to take into account any security
implications. Here you can find some recommendations to mitigate possible
vulnerabilities, such as Cross-Site Scripting (XSS):

1. If possible, serve user uploaded files from a separate domain
   (not a subdomain).

2. By default, Invenio-Files-REST sets some response headers to prevent
   the browser from rendering and executing HTML files.
   See :py:func:`invenio_files_rest.helpers.send_stream` for more information.

3. Prefer file download instead of allowing the browser to preview any file,
   by adding the :code:`?download` URL query argument


.. _usage-signals:

Signals
-------
Invenio-Files-REST supports signals that can be used to react to events.

Events are sent whenever a file is downloaded, uploaded or deleted.

As an example, let's listen to the file download event:

.. code-block:: python

    from invenio_files_rest.signals import file_downloaded

    def after_file_downloaded(event, sender_app, obj=None, **kwargs):
        print("File downloaded {0}".format(obj))

    listener = file_downloaded.connect(after_file_downloaded)
    # Request to download a file for the event to trigger

See :mod:`invenio_files_rest.signals` for more documentation.


.. _usage-integrity:

Integrity
^^^^^^^^^
Invenio-Files-REST computes and stores checksums when files are uploaded and it
allows you to set up periodic tasks to regularly re-validate files integrity.

By default, it uses :code:`MD5` to compute checksums. You can override this
by subclassing :py:class:`invenio_files_rest.storage.FileStorage`.

You can use the tasks :py:func:`invenio_files_rest.tasks.verify_checksum` and
:py:func:`invenio_files_rest.tasks.schedule_checksum_verification` to set up
periodic tasks to perform checksum verifications on single files or batches
and provide reports.

Let's create a periodic task to compute checksums:

.. code-block:: python

    CELERY_BEAT_SCHEDULE = {
        'file-checks': {
           'task': 'invenio_files_rest.tasks.schedule_checksum_verification',
           'schedule': timedelta(hours=1),
        }
    }

By default, :py:func:`invenio_files_rest.tasks.schedule_checksum_verification`
will generate batches of files to check using some predefined constraints,
in order to throttle the execution rate of the checks.
It will then spawn a celery task
:py:func:`invenio_files_rest.tasks.verify_checksum` for each of the file in
the set.

You can customize most of these parameters by passing the method arguments
to the schedule definition.

Keep in mind that you need to have :code:`celerybeat` running.


.. _usage-storage-backends:

Storage Backends
----------------
Invenio-Files-REST provides a default implementation of storage factory
:py:class:`invenio_files_rest.storage.PyFSFileStorage`
used when performing operation on files in the defined locations.
The :code:`PyFSFileStorage` class uses
`PyFilesystem <https://www.pyfilesystem.org/>`_ to access the file system.


Build your own Storage Backend
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
In order to use a different storage backend, you can implement the
:py:class:`invenio_files_rest.storage.FileStorage` interface.

Mandatory methods to implement:

* :code:`initialize`
* :code:`open`
* :code:`save`
* :code:`update`
* :code:`delete`

Optional methods to implement:

* :code:`send_file`
* :code:`checksum`
* :code:`copy`
* :code:`_init_hash`
* :code:`_compute_checksum`
* :code:`_write_stream`

Then, you will have to re-implement a storage factory in a similar way as the
default :py:func:`invenio_files_rest.storage.pyfs_storage_factory` and
set configuration variable
:py:data:`invenio_files_rest.config.FILES_REST_STORAGE_FACTORY`.


.. _usage-js-uploaders:

JS Uploaders
------------
Some JS uploaders do not allow you to customize the HTTP request that is
sent to the REST APIs when uploading a file. If the default implementation
provided by Invenio-Files-REST is not compatible, you will have to
implement your own custom factory to adapt the JS uploader request to
Invenio-Files-REST.

When using the AngularJS uploader
`ng-file-upload <https://github.com/danialfarid/ng-file-upload>`_,
Invenio-Files-REST already provides a compatible factory,
:py:func:`invenio_files_rest.views.ngfileupload_uploadfactory`.

If you have to create a new custom factory, you have to:

1. Create your own factory similar to
   :py:func:`invenio_files_rest.views.ngfileupload_uploadfactory`.

2. Instruct Invenio-Files-REST to use it by setting the configuration variables
:py:data:`invenio_files_rest.config.FILES_REST_MULTIPART_PART_FACTORIES` and
:py:data:`invenio_files_rest.config.FILES_REST_UPLOAD_FACTORIES`


.. _usage-multipart-upload:

Multipart Upload
----------------
You might want to optimize upload in case of large files.
Invenio-Files-REST allows you to upload parts of the same file in parallel
via multiparts uploads.

A multipart upload requires that each part of the file has the
same size, except for the last one that can be smaller.
Each part can be uploaded at the same
time and at the end of the process all parts are merged into one single file.

In case of failure when uploading one of the parts, the operation is completely
aborted and all parts are deleted.

With Invenio-Files-REST, the multipart upload consists of 3 actions:

* An initial request to initiate the upload and obtain an :code:`id`
  to be used for each part upload.

* A series of requests to upload of each part specifying
  the part number to correctly merge the file at the end.

* A final request to to merge all parts together.

Let's see an example. Let's create an 11 MB file which will then be split
into 2 chunks using the linux :code:`split` command:

.. code-block:: console

   $ dd if=/dev/urandom of=my_file.txt bs=1048576 count=11

   $ split -b6291456 my_file.txt segment_

Create a new bucket:

.. code-block:: console

   $ curl -X POST http://localhost:5000/api/files

Response:

.. code-block:: json

    {
       "max_file_size":null,
       "updated":"2019-05-17T06:52:52.897378+00:00",
       "locked":false,
       "links":{
          "self":"http://localhost:5000/api/files/
                  c896d17b-0e7d-44b3-beba-7e43b0b1a7a4",
          "uploads":"http://localhost:5000/api/files/
                     c896d17b-0e7d-44b3-beba-7e43b0b1a7a4?uploads",
          "versions":"http://localhost:5000/api/files/
                      c896d17b-0e7d-44b3-beba-7e43b0b1a7a4?versions"
       },
       "created":"2019-05-17T06:52:52.897373+00:00",
       "quota_size":null,
       "id":"c896d17b-0e7d-44b3-beba-7e43b0b1a7a4",
       "size":0
    }

Now, let's initiate the multipart upload. Notice the URL query argument
that specify total size and each part size:

.. code-block:: console

   $ B=c896d17b-0e7d-44b3-beba-7e43b0b1a7a4

   $ curl -i -X POST \
     "http://localhost:5000/api/files/$B/my_file.txt?
      uploads&size=11534336&partSize=6291456"

Notice the upload :code:`id` in the response:

.. code-block:: json

    {
       "updated":"2019-05-17T07:07:22.219002+00:00",
       "links":{
          "self":"http://localhost:5000/api/files/
                  c896d17b-0e7d-44b3-beba-7e43b0b1a7a4/my_file.txt?
                  uploadId=a85b1cbd-4080-4c81-a95c-b4df5d1b615f",

          "object":"http://localhost:5000/api/files/
                    c896d17b-0e7d-44b3-beba-7e43b0b1a7a4/my_file.txt",

          "bucket":"http://localhost:5000/api/files/
                    c896d17b-0e7d-44b3-beba-7e43b0b1a7a4"
       },
       "last_part_size":5242880,
       "created":"2019-05-17T07:07:22.218998+00:00",
       "bucket":"c896d17b-0e7d-44b3-beba-7e43b0b1a7a4",
       "completed":false,
       "part_size":6291456,
       "key":"my_file.txt",
       "last_part_number":1,
       "id":"a85b1cbd-4080-4c81-a95c-b4df5d1b615f",
       "size":11534336
    }

Now, let's upload each part in parallel. Notice the :code:`uploadId` and
:code:`partNumber` URL query arguments:

.. code-block:: console

   $ U=a85b1cbd-4080-4c81-a95c-b4df5d1b615f

   $ curl -i -X PUT --data-binary @segment_aa \
     "http://localhost:5000/api/files/$B/my_file.txt?uploadId=$U&partNumber=0"

   $ curl -i -X PUT --data-binary @segment_ab \
     "http://localhost:5000/api/files/$B/my_file.txt?uploadId=$U&partNumber=1"

Complete the multipart upload:

.. code-block:: console

   $ curl -i -X POST \
     "http://localhost:5000/api/files/$B/my_file.txt?uploadId=$U"

You can also abort a multipart upload (and delete all uploaded parts):

.. code-block:: console

   $ curl -i -X DELETE \
     "http://localhost:5000/api/files/$B/my_file.txt?uploadId=$U"

Multiparts uploads limits can be controlled via configuration variables:

* Set :py:data:`invenio_files_rest.config.FILES_REST_MULTIPART_MAX_PARTS`
  to limit the maximum number of parts for a single multipart upload.

* Set :py:data:`invenio_files_rest.config.FILES_REST_MULTIPART_CHUNKSIZE_MIN`
  to define the minimum size of each part.

* Set :py:data:`invenio_files_rest.config.FILES_REST_MULTIPART_CHUNKSIZE_MAX`
  to define the maximum size of each part.

* Set :py:data:`invenio_files_rest.config.FILES_REST_MULTIPART_EXPIRES`
  to define the maximum number of days for which a multipart upload is
  considered valid and accepts new part uploads.


.. _usage-large-files:

Large Files
-----------
By default, Flask and your web server have a limit on the maximum size of the
upload files. Normally, when the max size is exceeded, the server will return
a response code :code:`413 (Request Entity Too Large)`.

You can adjust these configurations according to your needs.

For Flask, specify :code:`MAX_CONTENT_LENGTH` configuration variable.
Be aware that if the request does not specify a :code:`CONTENT_LENGTH`,
no data will be read. To change the max size, you can for example:

.. code-block:: console

    $ app.config['MAX_CONTENT_LENGTH'] = 25 * 1024 * 1024

Here is an example for Nginx web server. If you are using another web server,
please check the related documentation.

.. code-block:: console

    http {
        ...
        client_max_body_size 25M;
    }


.. _usage-data-migration:

Data Migration
--------------
When you already have an instance running with a certain amount of uploaded
data, you might have the need to migrate the data to a different, larger or
more efficient physical location. It can involve your entire set of files or
just a part of it.

Note that files migration can be performed with no downtime and in a
completely transparent way for the user.

The steps to perform a complete migration are the followings:

1. Create the new :code:`Location` in the database with the URI of your
   new location and set it to :code:`default = True`. In this way, new
   :code:`Buckets` will use the new default location.
2. Change all existing buckets locations in the database to the new one.
   By doing this, any new file uploaded to the existing bucket will be stored
   in the new location.
3. For each :code:`FileInstance`, run the asynchronous task
   :py:func:`invenio_files_rest.tasks.migrate_file` passing the new location.

The asynchronous task :py:func:`invenio_files_rest.tasks.migrate_file`
will create a new :code:`FileInstance` and copy the file content to the
new location. It will then change each :code:`ObjectVersion` that have
a reference to the old :code:`FileInstance` to reference the new
:code:`FileInstance` and eventually run an integrity check.
"""

from .ext import InvenioFilesREST
from .proxies import current_files_rest

__version__ = "2.2.0"

__all__ = (
    "__version__",
    "current_files_rest",
    "InvenioFilesREST",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Implementation of various utility functions."""

import mimetypes

from flask import current_app
from werkzeug.utils import import_string

ENCODING_MIMETYPES = {
    "gzip": "application/gzip",
    "compress": "application/gzip",
    "bzip2": "application/x-bzip2",
    "xz": "application/x-xz",
}
"""Mapping encoding to MIME types which are not in mimetypes.types_map."""


def obj_or_import_string(value, default=None):
    """Import string or return object.

    :params value: Import path or class object to instantiate.
    :params default: Default object to return if the import fails.
    :returns: The imported object.
    """
    if isinstance(value, str):
        return import_string(value)
    elif value:
        return value
    return default


def load_or_import_from_config(key, app=None, default=None):
    """Load or import value from config.

    :returns: The loaded value.
    """
    app = app or current_app
    imp = app.config.get(key)
    return obj_or_import_string(imp, default=default)


def guess_mimetype(filename):
    """Map extra mimetype with the encoding provided.

    :returns: The extra mimetype.
    """
    m, encoding = mimetypes.guess_type(filename)
    if encoding:
        m = ENCODING_MIMETYPES.get(encoding, None)
    return m or "application/octet-stream"



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

import pytest
from flask import Flask, current_app

from invenio_files_rest import InvenioFilesREST


def test_version():
    """Test version import."""
    from invenio_files_rest import __version__

    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask("testapp")
    ext = InvenioFilesREST(app)
    assert "invenio-files-rest" in app.extensions

    app = Flask("testapp")
    ext = InvenioFilesREST()
    assert "invenio-files-rest" not in app.extensions
    ext.init_app(app)
    assert "invenio-files-rest" in app.extensions


def test_alembic(app, db):
    """Test alembic recipes."""
    ext = app.extensions["invenio-db"]

    if db.engine.name == "sqlite":
        raise pytest.skip("Upgrades are not supported on SQLite.")

    # skip index from alembic migrations until sqlalchemy 2.0
    # https://github.com/sqlalchemy/sqlalchemy/discussions/7597
    def include_object(object, name, type_, reflected, compare_to):
        if name == "ix_uq_partial_files_object_is_head":
            return False

        return True

    current_app.config["ALEMBIC_CONTEXT"] = {"include_object": include_object}

    assert not ext.alembic.compare_metadata()
    db.drop_all()
    ext.alembic.upgrade()

    assert not ext.alembic.compare_metadata()
    ext.alembic.stamp()
    ext.alembic.downgrade(target="96e796392533")
    ext.alembic.upgrade()

    assert not ext.alembic.compare_metadata()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create files REST tables."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = "2e97565eba72"
down_revision = "52ce868f33c3"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""

    def created():
        """Return instance of a column."""
        return sa.Column(
            "created",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        )

    def updated():
        """Return instance of a column."""
        return sa.Column(
            "updated",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        )

    def uri():
        """Return instance of a column."""
        return sa.Column(
            "uri", sa.Text().with_variant(mysql.VARCHAR(255), "mysql"), nullable=True
        )

    def key(nullable=True):
        """Return instance of a column."""
        return sa.Column(
            "key",
            sa.Text().with_variant(mysql.VARCHAR(255), "mysql"),
            nullable=nullable,
        )

    op.create_table(
        "files_files",
        created(),
        updated(),
        sa.Column("id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        uri(),
        sa.Column("storage_class", sa.String(length=1), nullable=True),
        sa.Column("size", sa.BigInteger(), nullable=True),
        sa.Column("checksum", sa.String(length=255), nullable=True),
        sa.Column("readable", sa.Boolean(name="readable"), nullable=False),
        sa.Column("writable", sa.Boolean(name="writable"), nullable=False),
        sa.Column("last_check_at", sa.DateTime(), nullable=True),
        sa.Column("last_check", sa.Boolean(name="last_check"), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("uri"),
    )
    op.create_table(
        "files_location",
        created(),
        updated(),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=20), nullable=False),
        sa.Column("uri", sa.String(length=255), nullable=False),
        sa.Column("default", sa.Boolean(name="default"), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_table(
        "files_bucket",
        created(),
        updated(),
        sa.Column("id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("default_location", sa.Integer(), nullable=False),
        sa.Column("default_storage_class", sa.String(length=1), nullable=False),
        sa.Column("size", sa.BigInteger(), nullable=False),
        sa.Column("quota_size", sa.BigInteger(), nullable=True),
        sa.Column("max_file_size", sa.BigInteger(), nullable=True),
        sa.Column("locked", sa.Boolean(name="locked"), nullable=False),
        sa.Column("deleted", sa.Boolean(name="deleted"), nullable=False),
        sa.ForeignKeyConstraint(
            ["default_location"], ["files_location.id"], ondelete="RESTRICT"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "files_buckettags",
        sa.Column("bucket_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("key", sa.String(length=255), nullable=False),
        sa.Column("value", sa.Text(), nullable=False),
        sa.ForeignKeyConstraint(["bucket_id"], ["files_bucket.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("bucket_id", "key"),
    )
    op.create_table(
        "files_multipartobject",
        created(),
        updated(),
        sa.Column("upload_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("bucket_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=True),
        key(),
        sa.Column("file_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("chunk_size", sa.Integer(), nullable=True),
        sa.Column("size", sa.BigInteger(), nullable=True),
        sa.Column("completed", sa.Boolean(name="completed"), nullable=False),
        sa.ForeignKeyConstraint(
            ["bucket_id"], ["files_bucket.id"], ondelete="RESTRICT"
        ),
        sa.ForeignKeyConstraint(["file_id"], ["files_files.id"], ondelete="RESTRICT"),
        sa.PrimaryKeyConstraint("upload_id"),
        sa.UniqueConstraint("upload_id", "bucket_id", "key", name="uix_item"),
    )
    op.create_table(
        "files_object",
        created(),
        updated(),
        sa.Column("bucket_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        key(nullable=False),
        sa.Column("version_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("file_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=True),
        sa.Column("_mimetype", sa.String(length=255), nullable=True),
        sa.Column("is_head", sa.Boolean(name="is_head"), nullable=False),
        sa.ForeignKeyConstraint(
            ["bucket_id"], ["files_bucket.id"], ondelete="RESTRICT"
        ),
        sa.ForeignKeyConstraint(["file_id"], ["files_files.id"], ondelete="RESTRICT"),
        sa.PrimaryKeyConstraint("bucket_id", "key", "version_id"),
    )
    op.create_index(
        op.f("ix_files_object__mimetype"), "files_object", ["_mimetype"], unique=False
    )
    op.create_table(
        "files_multipartobject_part",
        created(),
        updated(),
        sa.Column("upload_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column("part_number", sa.Integer(), autoincrement=False, nullable=False),
        sa.Column("checksum", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(
            ["upload_id"], ["files_multipartobject.upload_id"], ondelete="RESTRICT"
        ),
        sa.PrimaryKeyConstraint("upload_id", "part_number"),
    )


def downgrade():
    """Downgrade database."""
    op.drop_table("files_multipartobject_part")
    op.drop_index(op.f("ix_files_object__mimetype"), table_name="files_object")
    op.drop_table("files_object")
    op.drop_table("files_multipartobject")
    op.drop_table("files_buckettags")
    op.drop_table("files_bucket")
    op.drop_table("files_location")
    op.drop_table("files_files")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REST API serializers."""

import json
from time import sleep

from flask import current_app, request, url_for
from invenio_rest.serializer import BaseSchema as InvenioRestBaseSchema
from marshmallow import fields, missing, post_dump

from .errors import FilesException
from .models import Bucket, MultipartObject, ObjectVersion, Part


class BaseSchema(InvenioRestBaseSchema):
    """Base schema for all serializations."""

    created = fields.DateTime(dump_only=True)
    updated = fields.DateTime(dump_only=True)
    links = fields.Method("dump_links", dump_only=True)

    def dump_links(self, o):
        """Get base links."""
        return missing


class BucketSchema(BaseSchema):
    """Schema for bucket."""

    id = fields.UUID()
    size = fields.Integer()
    quota_size = fields.Integer()
    max_file_size = fields.Integer()
    locked = fields.Boolean()

    def dump_links(self, o):
        """Dump links."""
        return {
            "self": url_for(".bucket_api", bucket_id=o.id, _external=True),
            "versions": url_for(".bucket_api", bucket_id=o.id, _external=True)
            + "?versions",
            "uploads": url_for(".bucket_api", bucket_id=o.id, _external=True)
            + "?uploads",
        }


class ObjectVersionSchema(BaseSchema):
    """Schema for ObjectVersions."""

    key = fields.Str()
    version_id = fields.UUID()
    is_head = fields.Boolean()
    mimetype = fields.Str()
    size = fields.Integer(attribute="file.size")
    checksum = fields.String(attribute="file.checksum")
    delete_marker = fields.Boolean(attribute="deleted")
    tags = fields.Method("dump_tags", dump_only=True)

    def dump_tags(self, o):
        """Dump tags."""
        return o.get_tags()

    def dump_links(self, o):
        """Dump links."""
        params = {"versionId": o.version_id}
        data = {
            "self": url_for(
                ".object_api",
                bucket_id=o.bucket_id,
                key=o.key,
                _external=True,
                **(params if not o.is_head or o.deleted else {})
            ),
            "version": url_for(
                ".object_api",
                bucket_id=o.bucket_id,
                key=o.key,
                _external=True,
                **params
            ),
        }

        if o.is_head and not o.deleted:
            data.update(
                {
                    "uploads": url_for(
                        ".object_api", bucket_id=o.bucket_id, key=o.key, _external=True
                    )
                    + "?uploads",
                }
            )

        return data

    @post_dump(pass_many=True)
    def wrap(self, data, many):
        """Wrap response in envelope."""
        if not many:
            return data
        else:
            data = {"contents": data}
            bucket = self.context.get("bucket")
            if bucket:
                data.update(BucketSchema().dump(bucket).data)
            return data


class MultipartObjectSchema(BaseSchema):
    """Schema for ObjectVersions."""

    _endpoint = ".object_api"

    id = fields.UUID(attribute="upload_id", dump_only=True)
    bucket = fields.UUID(attribute="bucket_id", dump_only=True)
    key = fields.Str(dump_only=True)
    completed = fields.Boolean(dump_only=True)
    size = fields.Integer()
    part_size = fields.Integer(attribute="chunk_size")
    last_part_number = fields.Integer(dump_only=True)
    last_part_size = fields.Integer(dump_only=True)

    def dump_links(self, o):
        """Dump links."""
        links = {
            "self": url_for(
                ".object_api",
                bucket_id=o.bucket_id,
                key=o.key,
                uploadId=o.upload_id,
                _external=True,
            ),
            "object": url_for(
                ".object_api",
                bucket_id=o.bucket_id,
                key=o.key,
                _external=True,
            ),
        }

        version_id = self.context.get("object_version_id")
        if version_id:
            links.update(
                {
                    "object_version": url_for(
                        ".object_api",
                        bucket_id=o.bucket_id,
                        key=o.key,
                        versionId=version_id,
                        _external=True,
                    )
                }
            )

        bucket = self.context.get("bucket")
        if bucket:
            links.update(
                {
                    "bucket": url_for(
                        ".bucket_api",
                        bucket_id=o.bucket_id,
                        _external=True,
                    )
                }
            )

        return links


class PartSchema(BaseSchema):
    """Schema for parts."""

    part_number = fields.Integer()
    start_byte = fields.Integer()
    end_byte = fields.Integer()
    checksum = fields.Str()

    links = None

    @post_dump(pass_many=True)
    def wrap(self, data, many):
        """Wrap response in envelope."""
        if not many:
            return data
        else:
            data = {"parts": data}
            multipart = self.context.get("multipart")
            if multipart:
                data.update(
                    MultipartObjectSchema(context={"bucket": multipart.bucket})
                    .dump(multipart)
                    .data
                )
            return data


serializer_mapping = {
    Bucket: BucketSchema,
    ObjectVersion: ObjectVersionSchema,
    MultipartObject: MultipartObjectSchema,
    Part: PartSchema,
}


def schema_from_context(context, serializer_mapping):
    """Determine which schema to use."""
    item_class = context.get("class")
    return (
        serializer_mapping[item_class] if item_class else BaseSchema,
        context.get("many", False),
    )


def _format_args():
    if request and request.args.get("prettyprint"):
        return dict(
            indent=2,
            separators=(", ", ": "),
        )
    else:
        return dict(
            indent=None,
            separators=(",", ":"),
        )


def wait_for_taskresult(task_result, content, interval, max_rounds):
    """Get helper to wait for async task result to finish.

    The task will periodically send whitespace to prevent the connection from
    being closed.

    :param task_result: The async task to wait for.
    :param content: The content to return when the task is ready.
    :param interval: The duration of a sleep period before check again if the
        task is ready.
    :param max_rounds: The maximum number of intervals the function check
        before returning an Exception.
    :returns: An iterator on the content or a
        :class:`invenio_files_rest.errors.FilesException` exception if the
        timeout happened or the job failed.
    """
    assert max_rounds > 0

    def _whitespace_waiting():
        current = 0
        while current < max_rounds and current != -1:
            if task_result.ready():
                # Task is done and we return
                current = -1
                if task_result.successful():
                    yield content
                else:
                    yield FilesException(description="Job failed.").get_body()
            else:
                # Send whitespace to prevent connection from closing.
                current += 1
                sleep(interval)
                yield b" "

        # Timed-out reached
        if current == max_rounds:
            yield FilesException(description="Job timed out.").get_body()

    return _whitespace_waiting()


def json_serializer(
    data=None,
    code=200,
    headers=None,
    context=None,
    etag=None,
    task_result=None,
    serializer_mapping=serializer_mapping,
    view_name=None,
):
    """Build a json flask response using the given data.

    :param data: The data to serialize. (Default: ``None``)
    :param code: The HTTP status code. (Default: ``200``)
    :param headers: The HTTP headers to include. (Default: ``None``)
    :param context: The schema class context. (Default: ``None``)
    :param etag: The ETag header. (Default: ``None``)
    :param task_result: Optionally you can pass async task to wait for.
        (Default: ``None``)
    :param serializer_mapping: Optionally provide the serializer with a
        different mapping.
    :param view_name: Optionally push to the marshmallow context the view
        name prefix,usefull in case of multiple routes pointing to the same
        blueprint.
    :returns: A Flask response with json data.
    :rtype: :py:class:`flask.Response`
    """
    context = context or {}
    if view_name:
        context.update({"view_name": view_name})
    schema_class, many = schema_from_context(context, serializer_mapping)

    if data is not None:
        # Generate JSON response
        data = json.dumps(
            schema_class(context=context).dump(data, many=many).data, **_format_args()
        )

        interval = current_app.config["FILES_REST_TASK_WAIT_INTERVAL"]
        max_rounds = int(
            current_app.config["FILES_REST_TASK_WAIT_MAX_SECONDS"] // interval
        )

        response = current_app.response_class(
            # Stream response if waiting for task result.
            data
            if task_result is None
            else wait_for_taskresult(
                task_result,
                data,
                interval,
                max_rounds,
            ),
            mimetype="application/json",
        )
    else:
        response = current_app.response_class(mimetype="application/json")

    response.status_code = code
    if headers is not None:
        response.headers.extend(headers)

    if etag:
        response.set_etag(etag)

    return response



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Test CLI."""

import os

from invenio_files_rest.cli import files as cmd


def test_simple_workflow(app, db, tmpdir):
    """Run simple workflow."""
    runner = app.test_cli_runner()

    source = os.path.join(os.path.dirname(__file__), "fixtures", "source")

    # Create a location to use
    result = runner.invoke(
        cmd, ["location", "create", "tmp", "file://" + tmpdir.strpath, "--default"]
    )
    assert 0 == result.exit_code

    # Create the same location (check idempotent)
    result = runner.invoke(
        cmd, ["location", "create", "tmp", "file://" + tmpdir.strpath, "--default"]
    )
    assert 0 == result.exit_code
    assert "already exists" in result.output

    # Passing no subcommand should use default command 'create' (idempotent)
    result = runner.invoke(
        cmd, ["location", "tmp", "file://" + tmpdir.strpath, "--default"]
    )
    assert 0 == result.exit_code
    assert "already exists" in result.output

    # Create a second one as default to check tmp is not default anymore
    result = runner.invoke(
        cmd, ["location", "create", "aux", "file://" + tmpdir.strpath, "--default"]
    )
    assert 0 == result.exit_code

    # List locations and check the default is correct
    result = runner.invoke(cmd, ["location", "list"])
    assert 0 == result.exit_code

    created_locations = result.output.split("\n")

    # tmp is not default
    assert "tmp" in created_locations[0]
    assert "as default False" in created_locations[0]
    # aux is default
    assert "aux" in created_locations[1]
    assert "as default True" in created_locations[1]

    # Set tmp back as default
    result = runner.invoke(cmd, ["location", "set-default", "tmp"])
    assert 0 == result.exit_code

    # List locations and check the default is correct
    result = runner.invoke(cmd, ["location", "list"])
    assert 0 == result.exit_code

    created_locations = result.output.split("\n")
    # tmp is default
    assert "tmp" in created_locations[0]
    assert "as default True" in created_locations[0]
    # aux is not default
    assert "aux" in created_locations[1]
    assert "as default False" in created_locations[1]

    ##
    # Buckets
    ##

    result = runner.invoke(cmd, ["bucket", "touch"])
    assert 0 == result.exit_code
    bucket_id = result.output.split("\n")[0]

    # Specify a directory where 2 files have same content.
    result = runner.invoke(cmd, ["bucket", "cp", source, bucket_id, "--checksum"])
    assert 0 == result.exit_code

    # Specify a file.
    result = runner.invoke(cmd, ["bucket", "cp", __file__, bucket_id])
    assert 0 == result.exit_code
    prev_num_files = len(tmpdir.listdir())

    # No new file should be created.
    result = runner.invoke(cmd, ["bucket", "cp", __file__, bucket_id, "--checksum"])
    assert 0 == result.exit_code
    assert len(tmpdir.listdir()) == prev_num_files



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Proxy definitions."""

from flask import current_app
from werkzeug.local import LocalProxy

current_files_rest = LocalProxy(lambda: current_app.extensions["invenio-files-rest"])
"""Helper proxy to access files rest state object."""

current_permission_factory = LocalProxy(lambda: current_files_rest.permission_factory)
"""Helper proxy to access to the configured permission factory."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""File serving helpers for Files REST API."""

import hashlib
import mimetypes
import os
import unicodedata
import warnings
from time import time
from urllib.parse import urlsplit

from flask import current_app, make_response, request
from werkzeug.datastructures import Headers
from werkzeug.urls import url_quote
from werkzeug.wsgi import FileWrapper

MIMETYPE_TEXTFILES = {"readme"}

MIMETYPE_WHITELIST = {
    "audio/mpeg",
    "audio/ogg",
    "audio/wav",
    "audio/webm",
    "image/gif",
    "image/jpeg",
    "image/png",
    "image/tiff",
    "text/plain",
}
"""List of whitelisted MIME types.

.. warning::

   Do not add new types to this list unless you know what you are doing. You
   could potentially open up for XSS attacks.
"""

MIMETYPE_PLAINTEXT = {
    "application/javascript",
    "application/json",
    "application/xhtml+xml",
    "application/xml",
    "text/css",
    "text/csv",
    "text/html",
    "image/svg+xml",
}


def chunk_size_or_default(chunk_size):
    """Use default chunksize if not configured."""
    return chunk_size or 5 * 1024 * 1024  # 5MiB


def send_stream(
    stream,
    filename,
    size,
    mtime,
    mimetype=None,
    restricted=True,
    as_attachment=False,
    etag=None,
    content_md5=None,
    chunk_size=None,
    conditional=True,
    trusted=False,
):
    """Send the contents of a file to the client.

    .. warning::

        It is very easy to be exposed to Cross-Site Scripting (XSS) attacks if
        you serve user uploaded files. Here are some recommendations:

            1. Serve user uploaded files from a separate domain
               (not a subdomain). This way a malicious file can only attack
               other user uploaded files.
            2. Prevent the browser from rendering and executing HTML files (by
               setting ``trusted=False``).
            3. Force the browser to download the file as an attachment
               (``as_attachment=True``).

    :param stream: The file stream to send.
    :param filename: The file name.
    :param size: The file size.
    :param mtime: A Unix timestamp that represents last modified time (UTC).
    :param mimetype: The file mimetype. If ``None``, the module will try to
        guess. (Default: ``None``)
    :param restricted: If the file is not restricted, the module will set the
        cache-control. (Default: ``True``)
    :param as_attachment: If the file is an attachment. (Default: ``False``)
    :param etag: If defined, it will be set as HTTP E-Tag.
    :param content_md5: If defined, a HTTP Content-MD5 header will be set.
    :param chunk_size: The chunk size.
    :param conditional: Make the response conditional to the request.
        (Default: ``True``)
    :param trusted: Do not enable this option unless you know what you are
        doing. By default this function will send HTTP headers and MIME types
        that prevents your browser from rendering e.g. a HTML file which could
        contain a malicious script tag.
        (Default: ``False``)
    :returns: A Flask response instance.
    """
    chunk_size = chunk_size_or_default(chunk_size)

    # Guess mimetype from filename if not provided.
    if mimetype is None and filename:
        mimetype = mimetypes.guess_type(filename)[0]
    if mimetype is None:
        mimetype = "application/octet-stream"

    # Construct headers
    headers = Headers()
    headers["Content-Length"] = size
    if content_md5:
        headers["Content-MD5"] = content_md5

    if not trusted:
        # Sanitize MIME type
        mimetype = sanitize_mimetype(mimetype, filename=filename)
        # See https://www.owasp.org/index.php/OWASP_Secure_Headers_Project
        # Prevent JavaScript execution
        headers["Content-Security-Policy"] = "default-src 'none';"
        # Prevent MIME type sniffing for browser.
        headers["X-Content-Type-Options"] = "nosniff"
        # Prevent opening of downloaded file by IE
        headers["X-Download-Options"] = "noopen"
        # Prevent cross domain requests from Flash/Acrobat.
        headers["X-Permitted-Cross-Domain-Policies"] = "none"
        # Prevent files from being embedded in frame, iframe and object tags.
        headers["X-Frame-Options"] = "deny"
        # Enable XSS protection (IE, Chrome, Safari)
        headers["X-XSS-Protection"] = "1; mode=block"

    # Force Content-Disposition for application/octet-stream to prevent
    # Content-Type sniffing.
    if as_attachment or mimetype == "application/octet-stream":
        # See https://github.com/pallets/flask/commit/0049922f2e690a6d
        try:
            filenames = {"filename": filename.encode("latin-1")}
        except UnicodeEncodeError:
            filenames = {"filename*": "UTF-8''%s" % url_quote(filename)}
            encoded_filename = unicodedata.normalize("NFKD", filename).encode(
                "latin-1", "ignore"
            )
            if encoded_filename:
                filenames["filename"] = encoded_filename
        headers.add("Content-Disposition", "attachment", **filenames)
    else:
        headers.add("Content-Disposition", "inline")

    # Construct response object.
    rv = current_app.response_class(
        FileWrapper(stream, buffer_size=chunk_size),
        mimetype=mimetype,
        headers=headers,
        direct_passthrough=True,
    )

    # Set etag if defined
    if etag:
        rv.set_etag(etag)

    # Set last modified time
    if mtime is not None:
        rv.last_modified = int(mtime)

    # Set cache-control
    if not restricted:
        rv.cache_control.public = True
        cache_timeout = current_app.get_send_file_max_age(filename)
        if cache_timeout is not None:
            rv.cache_control.max_age = cache_timeout
            rv.expires = int(time() + cache_timeout)

    if conditional:
        rv = rv.make_conditional(request)

    return rv


def sanitize_mimetype(mimetype, filename=None):
    """Sanitize a MIME type so the browser does not render the file."""
    # Allow some few mime type like plain text, images and audio.
    if mimetype in MIMETYPE_WHITELIST:
        return mimetype
    # Rewrite HTML, JavaScript, CSS etc to text/plain.
    if mimetype in MIMETYPE_PLAINTEXT or (
        filename and filename.lower() in MIMETYPE_TEXTFILES
    ):
        return "text/plain"
    # Default
    return "application/octet-stream"


def make_path(base_uri, path, filename, path_dimensions, split_length):
    """Generate a path as base location for file instance.

    :param base_uri: The base URI.
    :param path: The relative path.
    :param path_dimensions: Number of chunks the path should be split into.
    :param split_length: The length of any chunk.
    :returns: A string representing the full path.
    """
    assert len(path) > path_dimensions * split_length

    uri_parts = []
    for i in range(path_dimensions):
        uri_parts.append(path[0:split_length])
        path = path[split_length:]
    uri_parts.append(path)
    uri_parts.append(filename)

    return os.path.join(base_uri, *uri_parts)


def compute_md5_checksum(stream, **kwargs):
    """Get helper method to compute MD5 checksum from a stream.

    :param stream: The input stream.
    :returns: The MD5 checksum.
    """
    return compute_checksum(stream, "md5", hashlib.md5(), **kwargs)


def compute_checksum(
    stream, algo, message_digest, chunk_size=None, progress_callback=None
):
    """Get helper method to compute checksum from a stream.

    :param stream: File-like object.
    :param algo: Identifier for checksum algorithm.
    :param messsage_digest: A message digest instance.
    :param chunk_size: Read at most size bytes from the file at a time.
    :param progress_callback: Function accepting one argument with number
        of bytes read. (Default: ``None``)
    :returns: The checksum.
    """
    chunk_size = chunk_size_or_default(chunk_size)

    bytes_read = 0
    while 1:
        chunk = stream.read(chunk_size)
        if not chunk:
            if progress_callback:
                progress_callback(bytes_read)
            break
        message_digest.update(chunk)
        bytes_read += len(chunk)
        if progress_callback:
            progress_callback(bytes_read)
    return "{0}:{1}".format(algo, message_digest.hexdigest())


def populate_from_path(bucket, source, checksum=True, key_prefix="", chunk_size=None):
    """Populate a ``bucket`` from all files in path.

    :param bucket: The bucket (instance or id) to create the object in.
    :param source: The file or directory path.
    :param checksum: If ``True`` then a MD5 checksum will be computed for each
        file. (Default: ``True``)
    :param key_prefix: The key prefix for the bucket.
    :param chunk_size: Chunk size to read from file.
    :returns: A iterator for all
        :class:`invenio_files_rest.models.ObjectVersion` instances.
    """
    from .models import FileInstance, ObjectVersion

    def create_file(key, path):
        """Create new ``ObjectVersion`` from path or existing ``FileInstance``.

        It checks MD5 checksum and size of existing ``FileInstance``s.
        """
        key = key_prefix + key

        if checksum:
            file_checksum = compute_md5_checksum(
                open(path, "rb"), chunk_size=chunk_size
            )
            file_instance = FileInstance.query.filter_by(
                checksum=file_checksum, size=os.path.getsize(path)
            ).first()
            if file_instance:
                return ObjectVersion.create(bucket, key, _file_id=file_instance.id)
        return ObjectVersion.create(bucket, key, stream=open(path, "rb"))

    if os.path.isfile(source):
        yield create_file(os.path.basename(source), source)
    else:
        for root, dirs, files in os.walk(source, topdown=False):
            for name in files:
                filename = os.path.join(root, name)
                assert filename.startswith(source)
                parts = [p for p in filename[len(source) :].split(os.sep) if p]
                yield create_file("/".join(parts), os.path.join(root, name))


def create_file_streaming_redirect_response(obj):
    """Redirect response generating function."""
    warnings.warn("This streaming does not support multiple storage backends.")
    response = make_response()
    redirect_url_base = "/user_files/"
    redirect_url_key = urlsplit(obj.file.uri).path
    response.headers["X-Accel-Redirect"] = redirect_url_base + redirect_url_key[1:]
    return response



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Models for Invenio-Files-REST."""

from blinker import Namespace

_signals = Namespace()

file_downloaded = _signals.signal("file-downloaded")
"""File downloaded signal.

Sent when a file is downloaded.
"""

file_uploaded = _signals.signal("file-uploaded")
"""File uploaded signal.

Sent when a file is uploaded.
"""

file_deleted = _signals.signal("file-deleted")
"""File deleted signal.

Sent when a file is deleted.
"""
